"""Protocol version 9 specific structs."""

from __future__ import annotations

import bellows.types.basic as basic
from bellows.types.struct import (  # noqa: F401
    EmberAesMmoHashContext,
    EmberApsFrame,
    EmberBindingTableEntry,
    EmberCurrentSecurityState,
    EmberGpAddress,
    EmberInitialSecurityState,
    EmberMulticastTableEntry,
    EmberNeighborTableEntry,
    EmberNetworkParameters,
    EmberRouteTableEntry,
    EmberTokenData,
    EmberTokenInfo,
    EmberTokTypeStackZllData,
    EmberTokTypeStackZllSecurity,
    EmberZigbeeNetwork,
    EmberZllAddressAssignment,
    EmberZllDeviceInfoRecord,
    EmberZllInitialSecurityState,
    EmberZllNetwork,
    EmberZllSecurityAlgorithmData,
    EzspStruct,
)

from . import named


class EmberBeaconData(EzspStruct):
    """Beacon data structure."""

    # The channel of the received beacon.
    channel: basic.uint8_t
    # The LQI of the received beacon.
    lqi: basic.uint8_t
    # The RSSI of the received beacon.
    rssi: basic.int8s
    # The depth of the received beacon.
    depth: basic.uint8_t
    # The network update ID of the received beacon.
    nwkUpdateId: basic.uint8_t
    # The power level of the received beacon. This field is valid only if the beacon is
    # an enhanced beacon
    power: basic.int8s
    # The TC connectivity and long uptime from capacity field.
    parentPriority: basic.int8s
    # The PAN ID of the received beacon.
    panId: named.EmberPanId
    # The extended PAN ID of the received beacon.
    extendedPanId: named.ExtendedPanId
    # The sender of the received beacon.
    sender: named.EmberNodeId
    # Whether or not the beacon is enhanced.
    enhanced: named.Bool
    # Whether the beacon is advertising permit join.
    permitJoin: named.Bool
    # Whether the beacon is advertising capacity.
    hasCapacity: named.Bool


class EmberBeaconIterator(EzspStruct):
    """Defines an iterator that is used to loop over cached beacons. Do not write to
    fields denoted as Private.
    """

    # The retrieved beacon.
    beacon: EmberBeaconData
    # (Private) The index of the retrieved beacon.
    index: basic.uint8_t


class EmberBeaconClassificationParams(EzspStruct):
    """The parameters related to beacon prioritization."""

    # The minimum RSSI value for receiving packets that is used in some beacon
    # prioritization algorithms.
    minRssiForReceivingPkts: basic.int8s
    # The beacon classification mask that identifies which beacon prioritization
    # algorithm to pick and defines the relevant parameters.
    beaconClassificationMask: basic.uint16_t


class EmberKeyStruct(EzspStruct):
    # A structure containing a key and its associated data.
    # A bitmask indicating the presence of data within the various fields
    # in the structure.
    bitmask: named.EmberKeyStructBitmask
    # The type of the key.
    type: named.EmberKeyType
    # The actual key data.
    key: named.KeyData
    # The outgoing frame counter associated with the key.
    outgoingFrameCounter: basic.uint32_t
    # The frame counter of the partner device associated with the key.
    incomingFrameCounter: basic.uint32_t
    # The sequence number associated with the key.
    sequenceNumber: basic.uint8_t
    # The IEEE address of the partner device also in possession of the key.
    partnerEUI64: named.EUI64

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[named.EmberKeyStruct, bytes]:
        # XXX: EZSPv9 serialization bug
        if len(data) == 24:
            return cls(), data[24:]

        return super().deserialize(data)


class EmberGpSinkListEntry(EzspStruct):
    # A sink list entry
    # The sink list type.
    type: basic.uint8_t
    # The EUI64 of the target sink.
    sinkEUI: named.EUI64
    # The short address of the target sink.
    sinkNodeId: named.EmberNodeId


class EmberGpProxyTableEntry(EzspStruct):
    """The internal representation of a proxy table entry."""

    # The link key to be used to secure this pairing link.
    securityLinkKey: named.KeyData
    # Internal status of the proxy table entry.
    status: named.EmberGpProxyTableEntryStatus
    # The tunneling options
    # (this contains both options and extendedOptions from the spec).
    options: basic.uint32_t
    # The addressing info of the GPD.
    gpd: EmberGpAddress
    # The assigned alias for the GPD.
    assignedAlias: named.EmberNodeId
    # The security options field.
    securityOptions: basic.uint8_t
    # The security frame counter of the GPD.
    gpdSecurityFrameCounter: named.EmberGpSecurityFrameCounter
    # The list of sinks (hardcoded to 2 which is the spec minimum).
    sinkList: basic.FixedList[EmberGpSinkListEntry, 2]
    # The groupcast radius.
    groupcastRadius: basic.uint8_t
    # The search counter
    searchCounter: basic.uint8_t


class EmberGpSinkTableEntry(EzspStruct):
    """The internal representation of a sink table entry."""

    # Internal status of the sink table entry
    status: named.EmberGpSinkTableEntryStatus
    # The tunneling options
    # (this contains both options and extendedOptions from the spec).
    options: basic.uint32_t
    # The addressing info of the GPD.
    gpd: EmberGpAddress
    # The device id for the GPD.
    deviceId: basic.uint8_t
    # The list of sinks (hardcoded to 2 which is the spec minimum).
    sinkList: basic.FixedList[EmberGpSinkListEntry, 2]
    # The assigned alias for the GPD.
    assignedAlias: named.EmberNodeId
    # The groupcast radius.
    groupcastRadius: basic.uint8_t
    # The security options field.
    securityOptions: basic.uint8_t
    # The security frame counter of the GPD.
    gpdSecurityFrameCounter: named.EmberGpSecurityFrameCounter
    # The key to use for GPD.
    gpdKey: named.KeyData


class EmberDutyCycleLimits(EzspStruct):
    """A structure containing duty cycle limit configurations.

    All limits are absolute, and are required to be as follows:
    suspLimit > critThresh > limitThresh
    For example:
    suspLimit = 250 (2.5%), critThresh = 180 (1.8%), limitThresh 100 (1.00%).
    """

    # The vendor identifier field shall contain the vendor identifier of the node.
    vendorId: basic.uint16_t
    # The vendor string field shall contain the vendor string of the node.
    vendorString: basic.FixedList[basic.uint8_t, 7]


class EmberPerDeviceDutyCycle(EzspStruct):
    """A structure containing per device overall duty cycle consumed

    up to the suspend limit).
    """

    # Node Id of device whose duty cycle is reported.
    nodeId: named.EmberNodeId
    # Amount of overall duty cycle consumed (up to suspend limit).
    dutyCycleConsumed: named.EmberDutyCycleHectoPct


class EmberTransientKeyData(EzspStruct):
    """The transient key data structure. Added in ver. 5. Revised in ver 8"""

    # The IEEE address paired with the transient link key.
    eui64: named.EUI64
    # The key data structure matching the transient key.
    keyData: named.KeyData
    # This bitmask indicates whether various fields in the structure contain valid data.
    bitmask: named.EmberKeyStructBitmask
    # The number of seconds remaining before the key is automatically timed out of the
    # transient key table.
    remainingTimeSeconds: basic.uint16_t


class EmberChildData(EzspStruct):
    """A structure containing a child node's data."""

    # The EUI64 of the child
    eui64: named.EUI64
    # The node type of the child
    type: named.EmberNodeType
    # The short address of the child
    id: named.EmberNodeId
    # The phy of the child
    phy: basic.uint8_t
    # The power of the child
    power: basic.uint8_t
    # The timeout of the child
    timeout: basic.uint8_t

    # The GPD's EUI64.
    # gpdIeeeAddress: named.EUI64
    # The GPD's source ID.
    # sourceId: basic.uint32_t



""""EZSP Protocol version 7 protocol handler."""
import logging

import voluptuous

import bellows.config

from . import commands, config, types as v7_types
from ..v6 import EZSPv6

LOGGER = logging.getLogger(__name__)


class EZSPv7(EZSPv6):
    """EZSP Version 7 Protocol version handler."""

    VERSION = 7
    COMMANDS = commands.COMMANDS
    SCHEMAS = {
        bellows.config.CONF_EZSP_CONFIG: voluptuous.Schema(config.EZSP_SCHEMA),
        bellows.config.CONF_EZSP_POLICIES: voluptuous.Schema(config.EZSP_POLICIES_SCH),
    }
    types = v7_types



from __future__ import annotations

from zigpy.types import (  # noqa: F401
    FixedList,
    List,
    LVBytes,
    LVList,
    bitmap8,
    bitmap16,
    enum8,
    enum16,
    enum32,
    int8s,
    int16s,
    int24s,
    int32s,
    int40s,
    int48s,
    int56s,
    int64s,
    uint8_t,
    uint16_t,
    uint24_t,
    uint32_t,
    uint40_t,
    uint48_t,
    uint56_t,
    uint64_t,
)


class LVBytes32(LVBytes):
    _prefix_length = 4



import asyncio
import contextlib
import functools
import logging

import click
import zigpy.config as zigpy_conf

import bellows.ezsp
import bellows.types as t

LOGGER = logging.getLogger(__name__)


class CSVParamType(click.ParamType):
    name = "comma separated integers"

    def __init__(self, min=None, max=None):
        self.intrange = click.IntRange(min, max)

    def convert(self, value, param, ctx):
        values = [self.intrange.convert(v, param, ctx) for v in value.split(",")]
        return values


class ZigbeeNodeParamType(click.ParamType):
    name = "colon separated hex bytes"

    def convert(self, value, param, ctx):
        if ":" not in value or len(value) != 23:
            self.fail("Node format should be a 8 byte hex string separated by ':'")
        return t.EUI64.convert(value)


def background(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(f(*args, **kwargs))

    return inner


def app(f, app_startup=True, extra_config=None):
    database_file = None
    application = None

    async def async_inner(ctx, *args, **kwargs):
        nonlocal database_file
        nonlocal application
        app_config = {
            zigpy_conf.CONF_DEVICE: {
                zigpy_conf.CONF_DEVICE_PATH: ctx.obj["device"],
                zigpy_conf.CONF_DEVICE_BAUDRATE: ctx.obj["baudrate"],
                zigpy_conf.CONF_FLOW_CONTROL: ctx.obj["flow_control"],
            },
            zigpy_conf.CONF_DATABASE: ctx.obj["database_file"],
        }
        if extra_config:
            app_config.update(extra_config)
        application = await setup_application(app_config, startup=app_startup)
        ctx.obj["app"] = application
        await f(ctx, *args, **kwargs)
        await asyncio.sleep(0.5)
        await application.shutdown()

    def shutdown():
        with contextlib.suppress(Exception):
            application._ezsp.close()

    @functools.wraps(f)
    def inner(*args, **kwargs):
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(async_inner(*args, **kwargs))
        except:  # noqa: E722
            # It seems that often errors like a message send will try to send
            # two messages, and not reading all of them will leave the NCP in
            # a bad state. This seems to mitigate this somewhat. Better way?
            loop.run_until_complete(asyncio.sleep(0.5))
            raise
        finally:
            shutdown()

    return inner


def print_cb(frame_name, response):
    click.echo(f"Callback: {frame_name} {response}")


def channel_mask(channels):
    mask = 0
    for channel in channels:
        if not (11 <= channel <= 26):
            raise click.BadOptionUsage("channels must be from 11 to 26")
        mask |= 1 << channel
    return mask


async def setup(dev, baudrate, cbh=None, configure=True):
    app_config = bellows.zigbee.application.ControllerApplication.SCHEMA(
        {
            zigpy_conf.CONF_DEVICE: {
                zigpy_conf.CONF_DEVICE_PATH: dev,
                zigpy_conf.CONF_DEVICE_BAUDRATE: baudrate,
                zigpy_conf.CONF_DEVICE_FLOW_CONTROL: zigpy_conf.CONF_DEVICE_FLOW_CONTROL_DEFAULT,
            }
        }
    )

    app = bellows.zigbee.application.ControllerApplication(app_config)
    await app.connect()

    if cbh:
        app._ezsp.add_callback(cbh)

    return app._ezsp


async def setup_application(app_config, startup=True):
    app_config = bellows.zigbee.application.ControllerApplication.SCHEMA(app_config)
    app = await bellows.zigbee.application.ControllerApplication.new(
        app_config, start_radio=startup
    )
    return app


def check(ret, message, expected=0):
    if ret == expected:
        return
    if isinstance(expected, list) and ret in expected:
        return
    raise click.ClickException(message)


async def network_init(s):
    v = await s.networkInit()
    check(
        v[0],
        f"Failure initializing network: {v[0]}",
        [0, t.EmberStatus.NOT_JOINED],
    )
    return v


def parse_epan(epan):
    """Parse a user specified extended PAN ID"""
    epan_list = [t.uint8_t(x, 16) for x in epan.split(":")]
    return t.FixedList[t.uint8_t, 8](epan_list)


async def basic_tc_permits(s):
    async def set_policy(policy, decision):
        v = await s.setPolicy(policy, decision)
        check(v[0], f"Failed to set policy {policy} to {decision}: {v[0]}")

    await set_policy(
        s.types.EzspPolicyId.TC_KEY_REQUEST_POLICY,
        s.types.EzspDecisionId.DENY_TC_KEY_REQUESTS,
    )
    await set_policy(
        s.types.EzspPolicyId.APP_KEY_REQUEST_POLICY,
        s.types.EzspDecisionId.ALLOW_APP_KEY_REQUESTS,
    )
    await set_policy(
        s.types.EzspPolicyId.TRUST_CENTER_POLICY,
        s.types.EzspDecisionId.ALLOW_PRECONFIGURED_KEY_JOINS,
    )


def get_device(app, node):
    if node not in app.devices:
        click.echo(f"Device {node} is not in the device database")
        return None

    return app.devices[node]


def get_endpoint(app, node, endpoint_id):
    dev = get_device(app, node)
    if dev is None:
        return (dev, None)

    if endpoint_id not in dev.endpoints:
        click.echo("Device %s has no endpoint %d" % (node, endpoint_id))
        return (dev, None)

    return (dev, dev.endpoints[endpoint_id])


def get_in_cluster(app, node, endpoint_id, cluster_id):
    dev, endpoint = get_endpoint(app, node, endpoint_id)
    if endpoint is None:
        return (dev, endpoint, None)

    if cluster_id not in endpoint.in_clusters:
        click.echo(
            "Device %s has no cluster %d on endpoint %d"
            % (node, cluster_id, endpoint_id)
        )
        return (dev, endpoint, None)

    return (dev, endpoint, endpoint.in_clusters[cluster_id])



from . import types as t
from ..v9.commands import COMMANDS as COMMANDS_v9

COMMANDS = {
    **COMMANDS_v9,
    # Use the correct `EmberChildData` object with the extra field
    "getChildData": (
        0x004A,
        (t.uint8_t,),
        (t.EmberStatus, t.EmberChildData),
    ),
    "setChildData": (
        0x00AC,
        (t.uint8_t, t.EmberChildData),
        (t.EmberStatus,),
    ),
}



""""EZSP Protocol version 11 protocol handler."""
import voluptuous as vol

import bellows.config

from . import commands, config, types as v11_types
from ..v10 import EZSPv10


class EZSPv11(EZSPv10):
    """EZSP Version 11 Protocol version handler."""

    VERSION = 11
    COMMANDS = commands.COMMANDS
    SCHEMAS = {
        bellows.config.CONF_EZSP_CONFIG: vol.Schema(config.EZSP_SCHEMA),
        bellows.config.CONF_EZSP_POLICIES: vol.Schema(config.EZSP_POLICIES_SCH),
    }
    types = v11_types



from __future__ import annotations

import dataclasses
import typing

import bellows.ezsp.v4.types as types_v4
import bellows.ezsp.v6.types as types_v6
import bellows.ezsp.v7.types as types_v7
import bellows.types as t


@dataclasses.dataclass(frozen=True)
class RuntimeConfig:
    config_id: t.enum8
    value: int
    minimum: bool = False


@dataclasses.dataclass(frozen=True)
class ValueConfig:
    value_id: t.enum8
    value: typing.Any


DEFAULT_CONFIG_COMMON = [
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT,
        value=7680,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_STACK_PROFILE,
        value=2,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_SUPPORTED_NETWORKS,
        value=1,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_MULTICAST_TABLE_SIZE,
        value=16,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE,
        value=2,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_SECURITY_LEVEL,
        value=5,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_ADDRESS_TABLE_SIZE,
        value=16,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_PAN_ID_CONFLICT_REPORT_THRESHOLD,
        value=2,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_KEY_TABLE_SIZE,
        value=4,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_MAX_END_DEVICE_CHILDREN,
        value=32,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_APPLICATION_ZDO_FLAGS,
        value=(
            t.EmberZdoConfigurationFlags.APP_RECEIVES_SUPPORTED_ZDO_REQUESTS
            | t.EmberZdoConfigurationFlags.APP_HANDLES_UNSUPPORTED_ZDO_REQUESTS
        ),
    ),
    # Must be set last
    RuntimeConfig(types_v4.EzspConfigId.CONFIG_PACKET_BUFFER_COUNT, value=0xFF),
]

DEFAULT_CONFIG_LEGACY = [
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_SOURCE_ROUTE_TABLE_SIZE,
        value=16,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT,
        value=60,
    ),
    RuntimeConfig(
        config_id=types_v4.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT,
        value=8,
    ),
] + DEFAULT_CONFIG_COMMON


DEFAULT_CONFIG_NEW = [
    RuntimeConfig(
        config_id=types_v6.EzspConfigId.CONFIG_SOURCE_ROUTE_TABLE_SIZE,
        value=200,
        minimum=True,
    ),
    RuntimeConfig(
        config_id=types_v6.EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT,
        value=8,
    ),
    RuntimeConfig(
        config_id=(
            types_v6.EzspConfigId.CONFIG_TC_REJOINS_USING_WELL_KNOWN_KEY_TIMEOUT_S
        ),
        value=90,
    ),
    ValueConfig(
        value_id=types_v7.EzspValueId.VALUE_FORCE_TX_AFTER_FAILED_CCA_ATTEMPTS,
        value=t.uint8_t(1),
    ),
] + DEFAULT_CONFIG_COMMON


DEFAULT_CONFIG = {
    4: DEFAULT_CONFIG_LEGACY,
    5: DEFAULT_CONFIG_LEGACY,
    6: DEFAULT_CONFIG_LEGACY,
    7: DEFAULT_CONFIG_NEW,
    8: DEFAULT_CONFIG_NEW,
    9: DEFAULT_CONFIG_NEW,
    10: DEFAULT_CONFIG_NEW,
    11: DEFAULT_CONFIG_NEW,
    12: DEFAULT_CONFIG_NEW,
    13: DEFAULT_CONFIG_NEW,
}



"""CLI commands which use the application framework"""

import asyncio
import binascii

import click
import zigpy.config
import zigpy.endpoint
import zigpy.exceptions

import bellows.config
import bellows.ezsp
import bellows.types as t
import bellows.zigbee.application

from . import opts, util
from .main import main


@main.command()
@opts.database_file
@opts.channel
@opts.extended_pan
@opts.pan
@opts.network_key
@opts.network_key_seq
@click.pass_context
def form(ctx, database, channel, pan_id, extended_pan_id, network_key, network_key_seq):
    """Form a new ZigBee network"""
    ctx.obj["database_file"] = database
    extra_config = {
        zigpy.config.CONF_NWK: {
            zigpy.config.CONF_NWK_CHANNEL: channel,
            zigpy.config.CONF_NWK_EXTENDED_PAN_ID: extended_pan_id,
            zigpy.config.CONF_NWK_PAN_ID: pan_id,
            zigpy.config.CONF_NWK_KEY: network_key,
            zigpy.config.CONF_NWK_KEY_SEQ: network_key_seq,
        }
    }

    async def inner(ctx):
        app = ctx.obj["app"]
        await app.startup(auto_form=True)

    return util.app(inner, app_startup=False, extra_config=extra_config)(ctx)


@main.command()
@opts.database_file
@opts.duration_s
@click.pass_context
def permit(ctx, database, duration_s):
    """Allow devices to join this ZigBee network"""
    ctx.obj["database_file"] = database

    async def inner(ctx):
        app = ctx.obj["app"]
        await app.permit(duration_s)

        click.echo(f"Joins are permitted for the next {duration_s}s...")
        await asyncio.sleep(duration_s + 1)
        click.echo("Done")

    return util.app(inner)(ctx)


@main.command()
@click.argument("node", type=util.ZigbeeNodeParamType())
@click.argument("code")
@opts.database_file
@opts.duration_s
@click.pass_context
def permit_with_key(ctx, database, duration_s, node, code):
    """Allow devices to join this ZigBee network using an install code"""
    ctx.obj["database_file"] = database
    code = binascii.unhexlify(code)

    async def inner(ctx):
        app = ctx.obj["app"]
        try:
            await app.permit_with_key(node, code, duration_s)

            click.echo(f"Joins are permitted for the next {duration_s}s...")
            await asyncio.sleep(duration_s + 1)
            click.echo("Done")
        except Exception as e:
            click.echo(e)

    return util.app(inner)(ctx)


@main.command()
@opts.database_file
@click.pass_context
def devices(ctx, database):
    """Show device database"""

    def print_clusters(title, clusters):
        clusters = sorted(clusters.items())
        if clusters:
            click.echo(f"      {title}:")
        for cluster_id, cluster in clusters:
            click.echo(f"        {cluster.name} ({cluster_id})")

    loop = asyncio.get_event_loop()
    config = {
        zigpy.config.CONF_DATABASE: database,
        bellows.config.CONF_DEVICE: {bellows.config.CONF_DEVICE_PATH: "/dev/null"},
    }
    config = bellows.config.CONFIG_SCHEMA(config)
    app = loop.run_until_complete(
        bellows.zigbee.application.ControllerApplication.new(config, start_radio=False)
    )
    for ieee, dev in app.devices.items():
        click.echo("Device:")
        click.echo(f"  NWK: 0x{dev.nwk:04x}")
        click.echo(f"  IEEE: {ieee}")
        click.echo("  Endpoints:")
        for epid, ep in dev.endpoints.items():
            if epid == 0:
                continue
            if ep.status == zigpy.endpoint.Status.NEW:
                click.echo("    %s: Uninitialized")
            else:
                click.echo(
                    "    %s: profile=0x%02x, device_type=%s"
                    % (epid, ep.profile_id, ep.device_type)
                )
                print_clusters("Input Clusters", ep.in_clusters)
                print_clusters("Output Clusters", ep.out_clusters)


@main.group()
@click.pass_context
@opts.database_file
@click.argument("node", type=util.ZigbeeNodeParamType())
def zdo(ctx, node, database):
    """Perform ZDO operations against a device"""
    ctx.obj["node"] = node
    ctx.obj["database_file"] = database


@zdo.command()
@click.pass_context
@util.app
async def endpoints(ctx):
    """List endpoints on a device"""
    app = ctx.obj["app"]
    node = ctx.obj["node"]

    dev = util.get_device(app, node)
    if dev is None:
        return

    try:
        v = await dev.zdo.request(0x0005, dev.nwk)
        if v[0] != t.EmberStatus.SUCCESS:
            click.echo(f"Non-success response: {v}")
        else:
            click.echo(v[2])
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@util.app
@click.argument("endpoint", type=click.IntRange(1, 255))
async def get_endpoint(ctx, endpoint):
    """Show an endpoint's simple descriptor"""
    app = ctx.obj["app"]
    node = ctx.obj["node"]

    dev, endp = util.get_endpoint(app, node, endpoint)
    if endp is None:
        return

    try:
        v = await dev.zdo.request(0x0004, dev.nwk, endpoint)
        if v[0] != t.EmberStatus.SUCCESS:
            click.echo(f"Non-success response: {v}")
        else:
            click.echo(v[2])
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@click.argument("endpoint", type=click.IntRange(1, 255))
@click.argument("cluster", type=click.IntRange(0, 65535))
@util.app
async def bind(ctx, endpoint, cluster):
    """Bind to a cluster on a node"""
    app = ctx.obj["app"]
    node = ctx.obj["node"]

    dev, endp, clust = util.get_in_cluster(app, node, endpoint, cluster)
    if clust is None:
        return

    try:
        v = await dev.zdo.bind(clust)
        click.echo(v)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@click.argument("endpoint", type=click.IntRange(1, 255))
@click.argument("cluster", type=click.IntRange(0, 65535))
@util.app
async def unbind(ctx, endpoint, cluster):
    """Unbind a cluster on a node"""
    app = ctx.obj["app"]
    node = ctx.obj["node"]

    dev, endp, clust = util.get_in_cluster(app, node, endpoint, cluster)
    if clust is None:
        return

    try:
        v = await dev.zdo.unbind(endpoint, cluster)
        click.echo(v)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zdo.command()
@click.pass_context
@util.app
async def leave(ctx):
    """Tell a node to leave the network"""
    app = ctx.obj["app"]

    try:
        v = await app.remove(ctx.obj["node"])
        click.echo(v)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@main.group()
@click.pass_context
@opts.database_file
@click.argument("node", type=util.ZigbeeNodeParamType())
@click.argument("endpoint", type=click.IntRange(1, 255))
@click.argument("cluster", type=click.IntRange(0, 65535))
def zcl(ctx, database, node, cluster, endpoint):
    """Perform ZCL operations against a device"""
    ctx.obj["database_file"] = database
    ctx.obj["node"] = node
    ctx.obj["endpoint"] = endpoint
    ctx.obj["cluster"] = cluster


@zcl.command()
@click.pass_context
@click.argument("attribute", type=click.IntRange(0, 65535))
@opts.manufacturer
@util.app
async def read_attribute(ctx, attribute, manufacturer):
    app = ctx.obj["app"]
    node = ctx.obj["node"]
    endpoint_id = ctx.obj["endpoint"]
    cluster_id = ctx.obj["cluster"]

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.read_attributes(
            [attribute], allow_cache=False, manufacturer=manufacturer
        )
        if not v:
            click.echo("Received empty response")
        elif attribute not in v[0]:
            click.echo(
                "Attribute {} not successful. Status={}".format(
                    attribute, v[1][attribute]
                )
            )
        else:
            click.echo(f"{attribute}={v[0][attribute]}")
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zcl.command()
@click.pass_context
@click.argument("attribute", type=click.IntRange(0, 65535))
@click.argument("value", type=click.IntRange(0, 65535))
@opts.manufacturer
@util.app
async def write_attribute(ctx, attribute, value, manufacturer):
    app = ctx.obj["app"]
    node = ctx.obj["node"]
    endpoint_id = ctx.obj["endpoint"]
    cluster_id = ctx.obj["cluster"]

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.write_attributes(
            {attribute: value}, manufacturer=manufacturer
        )
        click.echo(v)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zcl.command()
@click.pass_context
@util.app
async def commands(ctx):
    app = ctx.obj["app"]
    node = ctx.obj["node"]
    endpoint_id = ctx.obj["endpoint"]
    cluster_id = ctx.obj["cluster"]

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    for c in cluster.commands:
        click.echo(c)


@zcl.command()
@click.pass_context
@click.argument("command")
@click.argument("parameters", nargs=-1)
@opts.manufacturer
@util.app
async def command(ctx, command, parameters, manufacturer):
    app = ctx.obj["app"]
    node = ctx.obj["node"]
    endpoint_id = ctx.obj["endpoint"]
    cluster_id = ctx.obj["cluster"]

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await getattr(cluster, command)(*parameters, manufacturer=manufacturer)
        click.echo(v)
    except ValueError as e:
        click.echo(e)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)


@zcl.command()
@click.pass_context
@click.argument("attribute", type=click.IntRange(0, 65535))
@click.argument("min_interval", type=click.IntRange(0, 65535))
@click.argument("max_interval", type=click.IntRange(0, 65535))
@click.argument("reportable_change", type=click.INT)
@util.app
async def configure_reporting(
    ctx, attribute, min_interval, max_interval, reportable_change
):
    app = ctx.obj["app"]
    node = ctx.obj["node"]
    endpoint_id = ctx.obj["endpoint"]
    cluster_id = ctx.obj["cluster"]

    dev, endpoint, cluster = util.get_in_cluster(app, node, endpoint_id, cluster_id)
    if cluster is None:
        return

    try:
        v = await cluster.configure_reporting(
            attribute, min_interval, max_interval, reportable_change
        )
        click.echo(v)
    except zigpy.exceptions.ZigbeeException as e:
        click.echo(e)



"""Test network state repairs."""

import logging
from unittest.mock import AsyncMock, call

import pytest

from bellows.exception import InvalidCommandError
from bellows.ezsp import EZSP
from bellows.ezsp.v9.commands import GetTokenDataRsp
import bellows.types as t
from bellows.zigbee import repairs

from tests.test_ezsp import ezsp_f, make_ezsp


@pytest.fixture
def ezsp_tclk_f(ezsp_f: EZSP) -> EZSP:
    """Mock an EZSP instance with a valid TCLK."""
    ezsp_f.getEui64 = AsyncMock(
        return_value=[t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")]
    )
    ezsp_f.getTokenData = AsyncMock(side_effect=InvalidCommandError())
    ezsp_f.getCurrentSecurityState = AsyncMock(
        return_value=[
            t.EmberStatus.SUCCESS,
            t.EmberCurrentSecurityState(
                bitmask=(
                    t.EmberCurrentSecurityBitmask.GLOBAL_LINK_KEY
                    | t.EmberCurrentSecurityBitmask.HAVE_TRUST_CENTER_LINK_KEY
                    | 224
                ),
                trustCenterLongAddress=t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA"),
            ),
        ]
    )
    return ezsp_f


async def test_fix_invalid_tclk_noop(ezsp_tclk_f: EZSP, caplog) -> None:
    """Test that the TCLK is not rewritten unnecessarily."""

    ezsp_tclk_f.getEui64.return_value[0] = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")
    ezsp_tclk_f.getCurrentSecurityState.return_value[
        1
    ].trustCenterLongAddress = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")

    with caplog.at_level(logging.WARNING):
        assert await repairs.fix_invalid_tclk_partner_ieee(ezsp_tclk_f) is False

    assert "Fixing invalid TCLK" not in caplog.text


async def test_fix_invalid_tclk_old_firmware(ezsp_tclk_f: EZSP, caplog) -> None:
    """Test that the TCLK is not rewritten when the firmware is too old."""

    ezsp_tclk_f.getTokenData = AsyncMock(side_effect=InvalidCommandError())
    ezsp_tclk_f.getEui64.return_value[0] = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")
    ezsp_tclk_f.getCurrentSecurityState.return_value[
        1
    ].trustCenterLongAddress = t.EUI64.convert("BB:BB:BB:BB:BB:BB:BB:BB")

    with caplog.at_level(logging.WARNING):
        assert await repairs.fix_invalid_tclk_partner_ieee(ezsp_tclk_f) is False

    assert "Fixing invalid TCLK" in caplog.text
    assert "NV3 interface not available in this firmware" in caplog.text


async def test_fix_invalid_tclk(ezsp_tclk_f: EZSP, caplog) -> None:
    """Test that the TCLK is not rewritten when the firmware is too old."""

    ezsp_tclk_f.setTokenData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp_tclk_f.getTokenData = AsyncMock(
        return_value=GetTokenDataRsp(
            status=t.EmberStatus.SUCCESS,
            value=t.NV3StackTrustCenterToken(
                mode=228,
                eui64=t.EUI64.convert("BB:BB:BB:BB:BB:BB:BB:BB"),
                key=t.KeyData.convert(
                    "21:8e:df:b8:50:a0:4a:b6:8b:c6:10:25:bc:4e:93:6a"
                ),
            ).serialize(),
        )
    )
    ezsp_tclk_f.getEui64.return_value[0] = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")
    ezsp_tclk_f.getCurrentSecurityState.return_value[
        1
    ].trustCenterLongAddress = t.EUI64.convert("BB:BB:BB:BB:BB:BB:BB:BB")

    with caplog.at_level(logging.WARNING):
        assert await repairs.fix_invalid_tclk_partner_ieee(ezsp_tclk_f) is True

    assert "Fixing invalid TCLK" in caplog.text
    assert "NV3 interface not available in this firmware" not in caplog.text

    assert ezsp_tclk_f.setTokenData.mock_calls == [
        call(
            t.NV3KeyId.NVM3KEY_STACK_TRUST_CENTER,
            0,
            t.NV3StackTrustCenterToken(
                mode=228,
                eui64=t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA"),
                key=t.KeyData.convert(
                    "21:8e:df:b8:50:a0:4a:b6:8b:c6:10:25:bc:4e:93:6a"
                ),
            ).serialize(),
        )
    ]


@pytest.mark.parametrize("version", EZSP._BY_VERSION)
@pytest.mark.parametrize("has_library", [(True, False)])
async def test_fix_invalid_tclk_all_versions(
    version: int, has_library: bool, ezsp_tclk_f: EZSP, caplog
) -> None:
    """Test that the TCLK is fixed (or not) on all versions."""

    ezsp = await make_ezsp(version)
    fw_has_token_interface = hasattr(ezsp, "setTokenData")

    if fw_has_token_interface:
        ezsp.setTokenData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
        ezsp.getTokenData = AsyncMock(
            return_value=GetTokenDataRsp(
                status=t.EmberStatus.SUCCESS,
                value=t.NV3StackTrustCenterToken(
                    mode=228,
                    eui64=t.EUI64.convert("BB:BB:BB:BB:BB:BB:BB:BB"),
                    key=t.KeyData.convert(
                        "21:8e:df:b8:50:a0:4a:b6:8b:c6:10:25:bc:4e:93:6a"
                    ),
                ).serialize(),
            )
        )

    if not has_library:
        ezsp.setTokenData = AsyncMock(return_value=[t.EmberStatus.LIBRARY_NOT_LOADED])
        ezsp.getTokenData = AsyncMock(
            return_value=GetTokenDataRsp(status=t.EmberStatus.LIBRARY_NOT_LOADED)
        )

    ezsp.getEui64 = ezsp_tclk_f.getEui64
    ezsp.getCurrentSecurityState = ezsp_tclk_f.getCurrentSecurityState

    ezsp.getEui64.return_value[0] = t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")
    ezsp.getCurrentSecurityState.return_value[
        1
    ].trustCenterLongAddress = t.EUI64.convert("BB:BB:BB:BB:BB:BB:BB:BB")

    with caplog.at_level(logging.WARNING):
        assert (
            await repairs.fix_invalid_tclk_partner_ieee(ezsp) is fw_has_token_interface
        )

    assert "Fixing invalid TCLK" in caplog.text

    if fw_has_token_interface and has_library:
        assert "NV3 interface not available in this firmware" not in caplog.text

        assert ezsp.setTokenData.mock_calls == [
            call(
                t.NV3KeyId.NVM3KEY_STACK_TRUST_CENTER,
                0,
                t.NV3StackTrustCenterToken(
                    mode=228,
                    eui64=t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA"),
                    key=t.KeyData.convert(
                        "21:8e:df:b8:50:a0:4a:b6:8b:c6:10:25:bc:4e:93:6a"
                    ),
                ).serialize(),
            )
        ]
    else:
        assert "NV3 interface not available in this firmware" in caplog.text



import string

import pytest

import bellows.ezsp.v4.commands
import bellows.ezsp.v5.commands
import bellows.ezsp.v6.commands
import bellows.ezsp.v7.commands
import bellows.ezsp.v8.commands


@pytest.fixture(
    params=[
        bellows.ezsp.v4.commands,
        bellows.ezsp.v5.commands,
        bellows.ezsp.v6.commands,
        bellows.ezsp.v7.commands,
        bellows.ezsp.v8.commands,
    ]
)
def commands(request):
    """Return commands for all EZSP protocol versions."""
    yield request.param.COMMANDS


def test_names(commands):
    """Test that names of commands seem valid"""
    anum = string.ascii_letters + string.digits
    for command in commands.keys():
        assert all([c in anum for c in command]), command


def test_ids(commands):
    """Test that frame IDs seem valid"""
    seen = set()
    for command, (cmd_id, _, _) in commands.items():
        assert 0 <= cmd_id <= 255, command
        assert cmd_id not in seen
        seen.add(cmd_id)


def test_parms(commands):
    """Test that parameter descriptions seem valid"""
    for command, params in commands.items():
        assert isinstance(params[1], tuple), command
        assert isinstance(params[2], tuple), command


def test_handlers(commands):
    """Test that handler methods only have responses"""
    for command, params in commands.items():
        if not command.endswith("Handler"):
            continue
        assert len(params[1]) == 0, command
        assert len(params[2]) > 0, command



"""Coordinator state repairs."""

import logging

import zigpy.types

from bellows.exception import InvalidCommandError
from bellows.ezsp import EZSP
import bellows.types as t

LOGGER = logging.getLogger(__name__)


async def fix_invalid_tclk_partner_ieee(ezsp: EZSP) -> bool:
    """Fix invalid TCLK partner IEEE address."""
    (ieee,) = await ezsp.getEui64()
    ieee = zigpy.types.EUI64(ieee)

    (status, state) = await ezsp.getCurrentSecurityState()
    assert status == t.EmberStatus.SUCCESS

    if state.trustCenterLongAddress == ieee:
        return False

    LOGGER.warning(
        "Fixing invalid TCLK partner IEEE (%s => %s)",
        state.trustCenterLongAddress,
        ieee,
    )

    try:
        rsp = await ezsp.getTokenData(t.NV3KeyId.NVM3KEY_STACK_TRUST_CENTER, 0)
        assert rsp.status == t.EmberStatus.SUCCESS
    except (InvalidCommandError, AttributeError, AssertionError):
        LOGGER.warning("NV3 interface not available in this firmware, please upgrade!")
        return False

    token, remaining = t.NV3StackTrustCenterToken.deserialize(rsp.value)
    assert not remaining
    assert token.eui64 == state.trustCenterLongAddress

    (status,) = await ezsp.setTokenData(
        t.NV3KeyId.NVM3KEY_STACK_TRUST_CENTER,
        0,
        token.replace(eui64=ieee).serialize(),
    )
    assert status == t.EmberStatus.SUCCESS

    return True



from bellows.types import deserialize, serialize  # noqa: F401, F403
from bellows.types.basic import *  # noqa: F401,F403

from .named import *  # noqa: F401, F403
from .struct import *  # noqa: F401, F403



""""EZSP Protocol version 13 protocol handler."""
from __future__ import annotations

import voluptuous as vol

import bellows.config
import bellows.types as t

from . import commands, config, types as v13_types
from ..v12 import EZSPv12


class EZSPv13(EZSPv12):
    """EZSP Version 13 Protocol version handler."""

    VERSION = 13
    COMMANDS = commands.COMMANDS
    SCHEMAS = {
        bellows.config.CONF_EZSP_CONFIG: vol.Schema(config.EZSP_SCHEMA),
        bellows.config.CONF_EZSP_POLICIES: vol.Schema(config.EZSP_POLICIES_SCH),
    }
    types = v13_types

    async def add_transient_link_key(
        self, ieee: t.EUI64, key: t.KeyData
    ) -> t.EmberStatus:
        (status,) = await self.importTransientKey(
            ieee,
            key,
            v13_types.sl_zb_sec_man_flags_t.NONE,
        )

        return status



import pytest

import bellows.ezsp.v11

from .async_mock import AsyncMock, MagicMock, patch


@pytest.fixture
def ezsp_f():
    """EZSP v11 protocol handler."""
    return bellows.ezsp.v11.EZSPv11(MagicMock(), MagicMock())


def test_ezsp_frame(ezsp_f):
    ezsp_f._seq = 0x22
    data = ezsp_f._ezsp_frame("version", 11)
    assert data == b"\x22\x00\x01\x00\x00\x0b"


def test_ezsp_frame_rx(ezsp_f):
    """Test receiving a version frame."""
    ezsp_f(b"\x01\x01\x80\x00\x00\x01\x02\x34\x12")
    assert ezsp_f._handle_callback.call_count == 1
    assert ezsp_f._handle_callback.call_args[0][0] == "version"
    assert ezsp_f._handle_callback.call_args[0][1] == [0x01, 0x02, 0x1234]


async def test_pre_permit(ezsp_f):
    """Test pre permit."""
    p1 = patch.object(ezsp_f, "setPolicy", new=AsyncMock())
    p2 = patch.object(
        ezsp_f,
        "addTransientLinkKey",
        new=AsyncMock(return_value=[ezsp_f.types.EmberStatus.SUCCESS]),
    )
    with p1 as pre_permit_mock, p2 as tclk_mock:
        await ezsp_f.pre_permit(-1.9)
    assert pre_permit_mock.await_count == 2
    assert tclk_mock.await_count == 1


def test_command_frames(ezsp_f):
    """Test alphabetical list of frames matches the commands."""
    assert set(ezsp_f.COMMANDS) == set(command_frames)
    for name, frame_id in command_frames.items():
        assert ezsp_f.COMMANDS[name][0] == frame_id
        assert ezsp_f.COMMANDS_BY_ID[frame_id][0] == name


command_frames = {
    "addEndpoint": 0x0002,
    "addOrUpdateKeyTableEntry": 0x0066,
    "addTransientLinkKey": 0x00AF,
    "addressTableEntryIsActive": 0x005B,
    "aesEncrypt": 0x0094,
    "aesMmoHash": 0x006F,
    "becomeTrustCenter": 0x0077,
    "bindingIsActive": 0x002E,
    "bootloadTransmitCompleteHandler": 0x0093,
    "broadcastNetworkKeySwitch": 0x0074,
    "broadcastNextNetworkKey": 0x0073,
    "calculateSmacs": 0x009F,
    "calculateSmacs283k1": 0x00EA,
    "calculateSmacsHandler": 0x00A0,
    "calculateSmacsHandler283k1": 0x00EB,
    "callback": 0x0006,
    "incomingNetworkStatusHandler": 0x00C4,
    "childJoinHandler": 0x0023,
    "clearBindingTable": 0x002A,
    "clearKeyTable": 0x00B1,
    "clearStoredBeacons": 0x003C,
    "clearTemporaryDataMaybeStoreLinkKey": 0x00A1,
    "clearTemporaryDataMaybeStoreLinkKey283k1": 0x00EE,
    "clearTransientLinkKeys": 0x006B,
    "counterRolloverHandler": 0x00F2,
    "customFrame": 0x0047,
    "customFrameHandler": 0x0054,
    "dGpSend": 0x00C6,
    "dGpSentHandler": 0x00C7,
    "debugWrite": 0x0012,
    "delayTest": 0x009D,
    "deleteBinding": 0x002D,
    "dsaSign": 0x00A6,
    "dsaSignHandler": 0x00A7,
    "dsaVerify": 0x00A3,
    "dsaVerify283k1": 0x00B0,
    "dsaVerifyHandler": 0x0078,
    "dutyCycleHandler": 0x004D,
    "echo": 0x0081,
    "energyScanRequest": 0x009C,
    "energyScanResultHandler": 0x0048,
    "eraseKeyTableEntry": 0x0076,
    "findAndRejoinNetwork": 0x0021,
    "findKeyTableEntry": 0x0075,
    "findUnusedPanId": 0x00D3,
    "formNetwork": 0x001E,
    "generateCbkeKeys": 0x00A4,
    "generateCbkeKeys283k1": 0x00E8,
    "generateCbkeKeysHandler": 0x009E,
    "generateCbkeKeysHandler283k1": 0x00E9,
    "getAddressTableRemoteEui64": 0x005E,
    "getAddressTableRemoteNodeId": 0x005F,
    "getBinding": 0x002C,
    "getBeaconClassificationParams": 0x00F3,
    "getBindingRemoteNodeId": 0x002F,
    "getCertificate": 0x00A5,
    "getCertificate283k1": 0x00EC,
    "getChildData": 0x004A,
    "setChildData": 0x00AC,
    "getConfigurationValue": 0x0052,
    "getCurrentDutyCycle": 0x004C,
    "getCurrentSecurityState": 0x0069,
    "getEui64": 0x0026,
    "getDutyCycleLimits": 0x004B,
    "getDutyCycleState": 0x0035,
    "getExtendedTimeout": 0x007F,
    "getExtendedValue": 0x0003,
    "getFirstBeacon": 0x003D,
    "getKey": 0x006A,
    "getKeyTableEntry": 0x0071,
    "getLibraryStatus": 0x0001,
    "getLogicalChannel": 0x00BA,
    "getMfgToken": 0x000B,
    "getMulticastTableEntry": 0x0063,
    "getNeighbor": 0x0079,
    "getNeighborFrameCounter": 0x003E,
    "setNeighborFrameCounter": 0x00AD,
    "getNetworkParameters": 0x0028,
    "getNextBeacon": 0x0004,
    "getNodeId": 0x0027,
    "getNumStoredBeacons": 0x0008,
    "getParentChildParameters": 0x0029,
    "getParentClassificationEnabled": 0x00F0,
    "getPhyInterfaceCount": 0x00FC,
    "getPolicy": 0x0056,
    "getRadioParameters": 0x00FD,
    "setRadioIeee802154CcaMode": 0x0095,
    "getRandomNumber": 0x0049,
    "getRouteTableEntry": 0x007B,
    "getRoutingShortcutThreshold": 0x00D1,
    "getSecurityKeyStatus": 0x00CD,
    "getSourceRouteTableEntry": 0x00C1,
    "getSourceRouteTableFilledSize": 0x00C2,
    "getSourceRouteTableTotalSize": 0x00C3,
    "getStandaloneBootloaderVersionPlatMicroPhy": 0x0091,
    "getTimer": 0x004E,
    "getToken": 0x000A,
    "getTokenCount": 0x0100,
    "getTokenInfo": 0x0101,
    "getTokenData": 0x0102,
    "setTokenData": 0x0103,
    "resetNode": 0x0104,
    "getTransientKeyTableEntry": 0x006D,
    "getTransientLinkKey": 0x00CE,
    "getTrueRandomEntropySource": 0x004F,
    "getValue": 0x00AA,
    "getXncpInfo": 0x0013,
    "getZllPrimaryChannelMask": 0x00D9,
    "getZllSecondaryChannelMask": 0x00DA,
    "gpProxyTableGetEntry": 0x00C8,
    "gpProxyTableLookup": 0x00C0,
    "gpProxyTableProcessGpPairing": 0x00C9,
    "gpSinkTableClearAll": 0x00E2,
    "gpSinkTableFindOrAllocateEntry": 0x00E1,
    "gpSinkTableGetEntry": 0x00DD,
    "gpSinkTableInit": 0x0070,
    "gpSinkTableLookup": 0x00DE,
    "gpSinkTableRemoveEntry": 0x00E0,
    "gpSinkTableSetEntry": 0x00DF,
    "gpepIncomingMessageHandler": 0x00C5,
    "idConflictHandler": 0x007C,
    "incomingBootloadMessageHandler": 0x0092,
    "incomingManyToOneRouteRequestHandler": 0x007D,
    "incomingMessageHandler": 0x0045,
    "incomingRouteErrorHandler": 0x0080,
    "incomingRouteRecordHandler": 0x0059,
    "incomingSenderEui64Handler": 0x0062,
    "invalidCommand": 0x0058,
    "isHubConnected": 0x00E6,
    "isUpTimeLong": 0x00E5,
    "isZllNetwork": 0x00BE,
    "joinNetwork": 0x001F,
    "joinNetworkDirectly": 0x003B,
    "launchStandaloneBootloader": 0x008F,
    "leaveNetwork": 0x0020,
    "lookupEui64ByNodeId": 0x0061,
    "lookupNodeIdByEui64": 0x0060,
    "macFilterMatchMessageHandler": 0x0046,
    "macPassthroughMessageHandler": 0x0097,
    "maximumPayloadLength": 0x0033,
    "messageSentHandler": 0x003F,
    "mfglibEnd": 0x0084,
    "mfglibGetChannel": 0x008B,
    "mfglibGetPower": 0x008D,
    "mfglibRxHandler": 0x008E,
    "mfglibSendPacket": 0x0089,
    "mfglibSetChannel": 0x008A,
    "mfglibSetPower": 0x008C,
    "mfglibStart": 0x0083,
    "mfglibStartStream": 0x0087,
    "mfglibStartTone": 0x0085,
    "mfglibStopStream": 0x0088,
    "mfglibStopTone": 0x0086,
    "multiPhySetRadioChannel": 0x00FB,
    "multiPhySetRadioPower": 0x00FA,
    "multiPhyStart": 0x00F8,
    "multiPhyStop": 0x00F9,
    "neighborCount": 0x007A,
    "networkFoundHandler": 0x001B,
    "networkInit": 0x0017,
    "networkState": 0x0018,
    "noCallbacks": 0x0007,
    "nop": 0x0005,
    "permitJoining": 0x0022,
    "pollCompleteHandler": 0x0043,
    "pollForData": 0x0042,
    "pollHandler": 0x0044,
    "proxyBroadcast": 0x0037,
    "rawTransmitCompleteHandler": 0x0098,
    "readAndClearCounters": 0x0065,
    "readCounters": 0x00F1,
    "remoteDeleteBindingHandler": 0x0032,
    "remoteSetBindingHandler": 0x0031,
    "removeDevice": 0x00A8,
    "replaceAddressTableEntry": 0x0082,
    "requestLinkKey": 0x0014,
    "resetToFactoryDefaults": 0x00CC,
    "scanCompleteHandler": 0x001C,
    "sendBootloadMessage": 0x0090,
    "sendBroadcast": 0x0036,
    "sendLinkPowerDeltaRequest": 0x00F7,
    "sendManyToOneRouteRequest": 0x0041,
    "sendMulticast": 0x0038,
    "sendMulticastWithAlias": 0x003A,
    "sendPanIdUpdate": 0x0057,
    "sendRawMessage": 0x0096,
    "sendRawMessageExtended": 0x0051,
    "sendReply": 0x0039,
    "sendTrustCenterLinkKey": 0x0067,
    "sendUnicast": 0x0034,
    "setAddressTableRemoteEui64": 0x005C,
    "setAddressTableRemoteNodeId": 0x005D,
    "setBeaconClassificationParams": 0x00EF,
    "setBinding": 0x002B,
    "setBindingRemoteNodeId": 0x0030,
    "setBrokenRouteErrorCode": 0x0011,
    "setChildData": 0x00AC,
    "setConcentrator": 0x0010,
    "setConfigurationValue": 0x0053,
    "setDutyCycleLimitsInStack": 0x0040,
    "setExtendedTimeout": 0x007E,
    "setHubConnectivity": 0x00E4,
    "setInitialSecurityState": 0x0068,
    "setKeyTableEntry": 0x0072,
    "setLogicalAndRadioChannel": 0x00B9,
    "setLongUpTime": 0x00E3,
    "setMacPollFailureWaitTime": 0x00F4,
    "setManufacturerCode": 0x0015,
    "setMfgToken": 0x000C,
    "setMulticastTableEntry": 0x0064,
    "setNeighborFrameCounter": 0x00AD,
    "setParentClassificationEnabled": 0x00E7,
    "setPolicy": 0x0055,
    "setPowerDescriptor": 0x0016,
    "setPreinstalledCbkeData": 0x00A2,
    "setPreinstalledCbkeData283k1": 0x00ED,
    "setRadioChannel": 0x009A,
    "setRadioIeee802154CcaMode": 0x0095,
    "setRadioPower": 0x0099,
    "setRoutingShortcutThreshold": 0x00D0,
    "setSecurityKey": 0x00CA,
    "setSecurityParameters": 0x00CB,
    "setSourceRoute": 0x00AE,
    "setSourceRouteDiscoveryMode": 0x005A,
    "setTimer": 0x000E,
    "setToken": 0x0009,
    "setValue": 0x00AB,
    "setZllAdditionalState": 0x00D6,
    "setZllNodeType": 0x00D5,
    "setZllPrimaryChannelMask": 0x00DB,
    "setZllSecondaryChannelMask": 0x00DC,
    "stackStatusHandler": 0x0019,
    "stackTokenChangedHandler": 0x000D,
    "startScan": 0x001A,
    "stopScan": 0x001D,
    "switchNetworkKeyHandler": 0x006E,
    "timerHandler": 0x000F,
    "trustCenterJoinHandler": 0x0024,
    "unicastCurrentNetworkKey": 0x0050,
    "unicastNwkKeyUpdate": 0x00A9,
    "unusedPanIdFoundHandler": 0x00D2,
    "updateTcLinkKey": 0x006C,
    "version": 0x0000,
    "writeNodeData": 0x00FE,
    "zigbeeKeyEstablishmentHandler": 0x009B,
    "zllAddressAssignmentHandler": 0x00B8,
    "zllClearTokens": 0x0025,
    "zllGetTokens": 0x00BC,
    "zllNetworkFoundHandler": 0x00B6,
    "zllNetworkOps": 0x00B2,
    "zllOperationInProgress": 0x00D7,
    "zllRxOnWhenIdleGetActive": 0x00D8,
    "zllScanCompleteHandler": 0x00B7,
    "zllSetDataToken": 0x00BD,
    "zllSetInitialSecurityState": 0x00B3,
    "zllSetNonZllNetwork": 0x00BF,
    "zllSetRadioIdleMode": 0x00D4,
    "zllSetRxOnWhenIdle": 0x00B5,
    "zllSetSecurityStateWithoutKey": 0x00CF,
    "zllStartScan": 0x00B4,
    "zllTouchLinkTargetHandler": 0x00BB,
}



from bellows.types import KeyData

from . import types as t
from ..v11.commands import COMMANDS as COMMANDS_v11

COMMANDS = {
    **COMMANDS_v11,
    "readAttribute": (
        0x0108,
        tuple(
            {
                "endpoint": t.uint8_t,
                "cluster": t.uint16_t,
                "attributeId": t.uint16_t,
                "mask": t.uint8_t,
                "manufacturerCode": t.uint16_t,
            }.values()
        ),
        tuple(
            {
                "status": t.EmberStatus,
                "dataType": t.uint8_t,
                "data": t.LVBytes,
            }.values()
        ),
    ),
    "writeAttribute": (
        0x0109,
        tuple(
            {
                "endpoint": t.uint8_t,
                "cluster": t.uint16_t,
                "attributeId": t.uint16_t,
                "mask": t.uint8_t,
                "manufacturerCode": t.uint16_t,
                "overrideReadOnlyAndDataType": t.Bool,
                "justTest": t.Bool,
                "dataType": t.uint8_t,
                "data": t.LVBytes,
            }.values()
        ),
        tuple(
            {
                "status": t.EmberStatus,
            }.values()
        ),
    ),
    "setPassiveAckConfig": (
        0x0105,
        tuple(
            {
                "config": t.uint8_t,
                "minAcksNeeded": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "status": t.EmberStatus,
            }.values()
        ),
    ),
    "checkKeyContext": (
        0x0110,
        tuple(
            {
                "config": t.uint8_t,
                "minAcksNeeded": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "status": t.EmberStatus,
            }.values()
        ),
    ),
    "childId": (
        0x0106,
        tuple(
            {
                "childIndex": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "childId": t.EmberNodeId,
            }.values()
        ),
    ),
    "exportKey": (
        0x0114,
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
            }.values()
        ),
        tuple(
            {
                "key": KeyData,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportLinkKeyByEui": (
        0x010D,
        tuple(
            {
                "eui64": t.EUI64,
            }.values()
        ),
        tuple(
            {
                "plaintext_key": KeyData,
                "index": t.uint8_t,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportLinkKeyByIndex": (
        0x010F,
        tuple(
            {
                "index": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "eui64": t.EUI64,
                "plaintext_key": KeyData,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportTransientKeyByEui": (
        0x0113,
        tuple(
            {
                "eui64": t.EUI64,
            }.values()
        ),
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "plaintext_key": KeyData,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportTransientKeyByIndex": (
        0x0112,
        tuple(
            {
                "index": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "plaintext_key": KeyData,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "getApsKeyInfo": (
        0x010C,
        tuple(
            {
                "context_in": t.sl_zb_sec_man_context_t,
            }.values()
        ),
        tuple(
            {
                "eui": t.EUI64,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "gpSinkCommission": (
        0x010A,
        tuple(
            {
                "options": t.uint8_t,
                "gpmAddrForSecurity": t.uint16_t,
                "gpmAddrForPairing": t.uint16_t,
                "sinkEndpoint": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "status": t.EmberStatus,
            }.values()
        ),
    ),
    "gpTranslationTableClear": (
        0x010B,
        (),
        (),
    ),
    "importKey": (
        0x0115,
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "key": KeyData,
            }.values()
        ),
        tuple(
            {
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "importLinkKey": (
        0x010E,
        tuple(
            {
                "index": t.uint8_t,
                "address": t.EUI64,
                "key": KeyData,
            }.values()
        ),
        tuple(
            {
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "importTransientKey": (
        0x0111,
        tuple(
            {
                "eui64": t.EUI64,
                "plaintext_key": KeyData,
                "flags": t.sl_zb_sec_man_flags_t,
            }.values()
        ),
        tuple(
            {
                "status": t.sl_Status,
            }.values()
        ),
    ),
}



import voluptuous as vol

from bellows.config import cv_optional_int, cv_uint16, extend_vol_schema

from ..v4.config import EZSP_POLICIES_SHARED
from ..v7 import config as v7_config
from .types import EzspConfigId, EzspPolicyId

EZSP_SCHEMA = extend_vol_schema(
    v7_config.EZSP_SCHEMA,
    {
        vol.Optional(EzspConfigId.CONFIG_NEIGHBOR_TABLE_SIZE.name): cv_optional_int(
            min=8, max=26
        ),
        vol.Optional(EzspConfigId.CONFIG_MAX_END_DEVICE_CHILDREN.name): cv_optional_int(
            min=0, max=64
        ),
        vol.Optional(
            EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT.name, default=8
        ): cv_optional_int(min=0, max=14),
        vol.Optional(EzspConfigId.CONFIG_KEY_TABLE_SIZE.name): cv_optional_int(min=0),
        vol.Optional(
            EzspConfigId.CONFIG_TC_REJOINS_USING_WELL_KNOWN_KEY_TIMEOUT_S.name
        ): cv_optional_int(min=0),
    },
)

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



import voluptuous as vol

from bellows.config import cv_uint16, extend_vol_schema

from ..v4 import config as v4_config
from ..v4.config import EZSP_POLICIES_SHARED
from .types import EzspConfigId, EzspPolicyId  # noqa: F401

EZSP_SCHEMA = extend_vol_schema(v4_config.EZSP_SCHEMA, {})

del EZSP_SCHEMA[v4_config.EzspConfigId.CONFIG_BROADCAST_ALARM_DATA_SIZE.name]
del EZSP_SCHEMA[v4_config.EzspConfigId.CONFIG_UNICAST_ALARM_DATA_SIZE.name]

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



"""Typing helpers for bellows."""

from typing import TYPE_CHECKING

# pylint: disable=invalid-name
ControllerApplicationType = "ControllerApplication"
GatewayType = "Gateway"

if TYPE_CHECKING:
    import bellows.uart
    import bellows.zigbee.application

    GatewayType = bellows.uart.Gateway
    ControllerApplicationType = bellows.zigbee.application.ControllerApplication



import asyncio
import logging
import time

import click

from . import util
from .main import main

LOGGER = logging.getLogger(__name__)


@main.command()
@click.option(
    "-c", "--channel", type=click.IntRange(11, 26), metavar="CHANNEL", required=True
)
@click.option(
    "-p", "--power", type=click.IntRange(-100, 20), metavar="POWER", required=True
)
@click.pass_context
def stream(ctx, channel, power):
    """Transmit random stream of characters on CHANNEL with POWER (in dBm)."""
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(_stream(ctx, channel, power))
    except KeyboardInterrupt:
        start_time = ctx.obj.get("start_time", None)
        if start_time:
            duration = time.time() - start_time
            click.echo(
                "\nStreamed on channel %d for %0.2fs" % (channel, duration), err=True
            )
    finally:
        if "ezsp" in ctx.obj:
            s = ctx.obj["ezsp"]
            loop.run_until_complete(s.mfglibStopStream())
            loop.run_until_complete(s.mfglibEnd())
            s.close()


async def _stream(ctx, channel, power):
    s = await util.setup(ctx.obj["device"], ctx.obj["baudrate"])
    ctx.obj["ezsp"] = s

    v = await s.mfglibStart(False)
    util.check(v[0], "Unable to start mfglib")

    v = await s.mfglibSetChannel(channel)
    util.check(v[0], "Unable to set channel")

    v = await s.mfglibSetPower(0, power)
    util.check(v[0], "Unable to set power")

    v = await s.mfglibStartStream()
    util.check(v[0], "Unable to start stream")
    click.echo("Started transmitting random stream of characters", err=True)
    ctx.obj["start_time"] = time.time()
    while True:
        await asyncio.sleep(3600)



import asyncio
import threading

import pytest
import serial_asyncio
import zigpy.config as conf

from bellows import uart
import bellows.types as t

from .async_mock import AsyncMock, MagicMock, call, patch, sentinel


@pytest.mark.parametrize("flow_control", ["software", "hardware"])
async def test_connect(flow_control, monkeypatch):
    appmock = MagicMock()
    transport = MagicMock()

    async def mockconnect(loop, protocol_factory, **kwargs):
        protocol = protocol_factory()
        loop.call_soon(protocol.connection_made, transport)
        return None, protocol

    monkeypatch.setattr(serial_asyncio, "create_serial_connection", mockconnect)
    gw = await uart.connect(
        conf.SCHEMA_DEVICE(
            {
                conf.CONF_DEVICE_PATH: "/dev/serial",
                conf.CONF_DEVICE_BAUDRATE: 115200,
                conf.CONF_DEVICE_FLOW_CONTROL: flow_control,
            }
        ),
        appmock,
        use_thread=False,
    )

    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0
    gw.close()


async def test_connect_threaded(monkeypatch):
    appmock = MagicMock()
    transport = MagicMock()

    async def mockconnect(loop, protocol_factory, **kwargs):
        protocol = protocol_factory()
        loop.call_soon(protocol.connection_made, transport)
        return None, protocol

    monkeypatch.setattr(serial_asyncio, "create_serial_connection", mockconnect)

    def on_transport_close():
        gw.connection_lost(None)

    transport.close.side_effect = on_transport_close
    gw = await uart.connect(
        conf.SCHEMA_DEVICE(
            {conf.CONF_DEVICE_PATH: "/dev/serial", conf.CONF_DEVICE_BAUDRATE: 115200}
        ),
        appmock,
    )

    # Need to close to release thread
    gw.close()

    # Ensure all threads are cleaned up
    [t.join(1) for t in threading.enumerate() if "bellows" in t.name]
    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0


async def test_connect_threaded_failure(monkeypatch):
    appmock = MagicMock()
    transport = MagicMock()

    mockconnect = AsyncMock()
    mockconnect.side_effect = OSError

    monkeypatch.setattr(serial_asyncio, "create_serial_connection", mockconnect)

    def on_transport_close():
        gw.connection_lost(None)

    transport.close.side_effect = on_transport_close
    with pytest.raises(OSError):
        gw = await uart.connect(
            conf.SCHEMA_DEVICE(
                {
                    conf.CONF_DEVICE_PATH: "/dev/serial",
                    conf.CONF_DEVICE_BAUDRATE: 115200,
                }
            ),
            appmock,
        )

    # Ensure all threads are cleaned up
    [t.join(1) for t in threading.enumerate() if "bellows" in t.name]
    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0


async def test_connect_threaded_failure_cancellation_propagation(monkeypatch):
    appmock = MagicMock()

    async def mock_connect(loop, protocol_factory, *args, **kwargs):
        protocol = protocol_factory()
        transport = AsyncMock()

        protocol.connection_made(transport)

        return transport, protocol

    with patch("bellows.uart.zigpy.serial.create_serial_connection", mock_connect):
        gw = await uart.connect(
            conf.SCHEMA_DEVICE(
                {
                    conf.CONF_DEVICE_PATH: "/dev/serial",
                    conf.CONF_DEVICE_BAUDRATE: 115200,
                }
            ),
            appmock,
            use_thread=True,
        )

    # Begin waiting for the startup reset
    wait_for_reset = gw.wait_for_startup_reset()

    # But lose connection halfway through
    asyncio.get_running_loop().call_later(0.1, gw.connection_lost, RuntimeError())

    # Cancellation should propagate to the outer loop
    with pytest.raises(RuntimeError):
        await wait_for_reset

    # Ensure all threads are cleaned up
    [t.join(1) for t in threading.enumerate() if "bellows" in t.name]
    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0


@pytest.fixture
def gw():
    gw = uart.Gateway(MagicMock())
    gw._transport = MagicMock()
    return gw


def test_close(gw):
    gw.close()
    assert gw._transport.close.call_count == 1


async def test_reset_timeout(gw, monkeypatch):
    monkeypatch.setattr(uart, "RESET_TIMEOUT", 0.1)
    with pytest.raises(asyncio.TimeoutError):
        await gw.reset()


async def test_reset_old(gw):
    future = asyncio.get_event_loop().create_future()
    future.set_result(sentinel.result)
    gw._reset_future = future
    ret = await gw.reset()
    assert ret == sentinel.result
    gw._transport.write.assert_not_called()


def test_connection_lost_exc(gw):
    gw.connection_lost(sentinel.exception)

    conn_lost = gw._application.connection_lost
    assert conn_lost.call_count == 1
    assert conn_lost.call_args[0][0] is sentinel.exception


def test_connection_closed(gw):
    gw.connection_lost(None)

    assert gw._application.connection_lost.call_count == 0


def test_eof_received(gw):
    gw.eof_received()

    assert gw._application.connection_lost.call_count == 1


async def test_connection_lost_reset_error_propagation(monkeypatch):
    app = MagicMock()
    transport = MagicMock()

    async def mockconnect(loop, protocol_factory, **kwargs):
        protocol = protocol_factory()
        loop.call_soon(protocol.connection_made, transport)
        return None, protocol

    monkeypatch.setattr(serial_asyncio, "create_serial_connection", mockconnect)

    def on_transport_close():
        gw.connection_lost(None)

    transport.close.side_effect = on_transport_close
    gw = await uart.connect(
        conf.SCHEMA_DEVICE(
            {conf.CONF_DEVICE_PATH: "/dev/serial", conf.CONF_DEVICE_BAUDRATE: 115200}
        ),
        app,
        use_thread=False,  # required until #484 is merged
    )

    asyncio.get_running_loop().call_later(0.1, gw.connection_lost, ValueError())

    with pytest.raises(ValueError):
        await gw.reset()

    # Need to close to release thread
    gw.close()

    # Ensure all threads are cleaned up
    [t.join(1) for t in threading.enumerate() if "bellows" in t.name]
    threads = [t for t in threading.enumerate() if "bellows" in t.name]
    assert len(threads) == 0


async def test_wait_for_startup_reset(gw):
    loop = asyncio.get_running_loop()
    loop.call_later(0.01, gw.reset_received, t.NcpResetCode.RESET_SOFTWARE)

    assert gw._startup_reset_future is None
    await gw.wait_for_startup_reset()
    assert gw._startup_reset_future is None


async def test_wait_for_startup_reset_failure(gw):
    assert gw._startup_reset_future is None

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(gw.wait_for_startup_reset(), 0.01)

    assert gw._startup_reset_future is None


async def test_callbacks(gw):
    gw.data_received(b"some ezsp packet")
    assert gw._application.frame_received.mock_calls == [call(b"some ezsp packet")]

    gw.error_received(t.NcpResetCode.RESET_SOFTWARE)
    assert gw._application.enter_failed_state.mock_calls == [
        call(t.NcpResetCode.RESET_SOFTWARE)
    ]



from __future__ import annotations

import abc
import asyncio
import binascii
import functools
import logging
import sys
from typing import TYPE_CHECKING, Any, Callable

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

from bellows.config import CONF_EZSP_POLICIES
from bellows.exception import InvalidCommandError
import bellows.types as t

if TYPE_CHECKING:
    from bellows.uart import Gateway

LOGGER = logging.getLogger(__name__)
EZSP_CMD_TIMEOUT = 6  # Sum of all ASH retry timeouts: 0.4 + 0.8 + 1.6 + 3.2


class ProtocolHandler(abc.ABC):
    """EZSP protocol specific handler."""

    COMMANDS = {}
    VERSION = None

    def __init__(self, cb_handler: Callable, gateway: Gateway) -> None:
        self._handle_callback = cb_handler
        self._awaiting = {}
        self._gw = gateway
        self._seq = 0
        self.COMMANDS_BY_ID = {
            cmd_id: (name, tx_schema, rx_schema)
            for name, (cmd_id, tx_schema, rx_schema) in self.COMMANDS.items()
        }
        self.tc_policy = 0

    def _ezsp_frame(self, name: str, *args: tuple[Any, ...]) -> bytes:
        """Serialize the named frame and data."""
        c = self.COMMANDS[name]
        frame = self._ezsp_frame_tx(name)
        data = self.types.serialize(args, c[1])
        return frame + data

    @abc.abstractmethod
    def _ezsp_frame_rx(self, data: bytes) -> tuple[int, int, bytes]:
        """Handler for received data frame."""

    @abc.abstractmethod
    def _ezsp_frame_tx(self, name: str) -> bytes:
        """Serialize the named frame."""

    async def pre_permit(self, time_s: int) -> None:
        """Schedule task before allowing new joins."""

    async def add_transient_link_key(
        self, ieee: t.EUI64, key: t.KeyData
    ) -> t.EmberStatus:
        """Add a transient link key."""

    async def command(self, name, *args) -> Any:
        """Serialize command and send it."""
        LOGGER.debug("Sending command  %s: %s", name, args)
        data = self._ezsp_frame(name, *args)
        cmd_id, _, rx_schema = self.COMMANDS[name]
        future = asyncio.get_running_loop().create_future()
        self._awaiting[self._seq] = (cmd_id, rx_schema, future)
        self._seq = (self._seq + 1) % 256

        async with asyncio_timeout(EZSP_CMD_TIMEOUT):
            await self._gw.send_data(data)
            return await future

    async def update_policies(self, policy_config: dict) -> None:
        """Set up the policies for what the NCP should do."""

        policies = self.SCHEMAS[CONF_EZSP_POLICIES](policy_config)
        self.tc_policy = policies[self.types.EzspPolicyId.TRUST_CENTER_POLICY.name]

        for policy, value in policies.items():
            (status,) = await self.setPolicy(self.types.EzspPolicyId[policy], value)
            assert status == self.types.EmberStatus.SUCCESS  # TODO: Better check

    def __call__(self, data: bytes) -> None:
        """Handler for received data frame."""
        orig_data = data
        sequence, frame_id, data = self._ezsp_frame_rx(data)

        try:
            frame_name, _, rx_schema = self.COMMANDS_BY_ID[frame_id]
        except KeyError:
            LOGGER.warning(
                "Unknown application frame 0x%04X received: %s (%s).  This is a bug!",
                frame_id,
                binascii.hexlify(data),
                binascii.hexlify(orig_data),
            )
            return

        try:
            if isinstance(rx_schema, tuple):
                result, data = self.types.deserialize(data, rx_schema)
            else:
                result, data = rx_schema.deserialize(data)
        except Exception:
            LOGGER.warning(
                "Failed to parse frame %s: %s", frame_name, binascii.hexlify(data)
            )
            raise

        LOGGER.debug("Received command %s: %s", frame_name, result)

        if data:
            LOGGER.debug("Frame contains trailing data: %s", data)

        if sequence in self._awaiting:
            expected_id, schema, future = self._awaiting.pop(sequence)
            try:
                if frame_name == "invalidCommand":
                    sent_cmd_name = self.COMMANDS_BY_ID[expected_id][0]
                    future.set_exception(
                        InvalidCommandError(
                            f"{sent_cmd_name} command is an {frame_name}, was sent "
                            f"under {sequence} sequence number: {result[0].name}"
                        )
                    )
                    return

                assert expected_id == frame_id
                future.set_result(result)
            except asyncio.InvalidStateError:
                LOGGER.debug(
                    "Error processing %s response. %s command timed out?",
                    sequence,
                    self.COMMANDS_BY_ID.get(expected_id, [expected_id])[0],
                )
        else:
            self._handle_callback(frame_name, result)

    def __getattr__(self, name: str) -> Callable:
        if name not in self.COMMANDS:
            raise AttributeError(f"{name} not found in COMMANDS")

        return functools.partial(self.command, name)



from .basic import *  # noqa: F401,F403
from .named import *  # noqa: F401,F403
from .struct import *  # noqa: F401,F403


def deserialize(data, schema):
    result = []
    for type_ in schema:
        value, data = type_.deserialize(data)
        result.append(value)
    return result, data


def serialize(data, schema):
    return b"".join(t(v).serialize() for t, v in zip(schema, data))



""""EZSP Protocol version 8 protocol handler."""
import asyncio
import logging
from typing import Tuple

import voluptuous

import bellows.config

from . import commands, config, types as v8_types
from ..v7 import EZSPv7

LOGGER = logging.getLogger(__name__)


class EZSPv8(EZSPv7):
    """EZSP Version 8 Protocol version handler."""

    VERSION = 8
    COMMANDS = commands.COMMANDS
    SCHEMAS = {
        bellows.config.CONF_EZSP_CONFIG: voluptuous.Schema(config.EZSP_SCHEMA),
        bellows.config.CONF_EZSP_POLICIES: voluptuous.Schema(config.EZSP_POLICIES_SCH),
    }
    types = v8_types

    def _ezsp_frame_tx(self, name: str) -> bytes:
        """Serialize the frame id."""
        cmd_id = self.COMMANDS[name][0]
        hdr = [self._seq, 0x00, 0x01]
        return bytes(hdr) + self.types.uint16_t(cmd_id).serialize()

    def _ezsp_frame_rx(self, data: bytes) -> Tuple[int, int, bytes]:
        """Handler for received data frame."""
        seq, data = data[0], data[3:]
        frame_id, data = self.types.uint16_t.deserialize(data)

        return seq, frame_id, data

    async def pre_permit(self, time_s: int) -> None:
        """Temporarily change TC policy while allowing new joins."""
        await super().pre_permit(time_s)
        await self.setPolicy(
            v8_types.EzspPolicyId.TRUST_CENTER_POLICY,
            v8_types.EzspDecisionBitmask.ALLOW_JOINS
            | v8_types.EzspDecisionBitmask.ALLOW_UNSECURED_REJOINS,
        )
        await asyncio.sleep(time_s + 2)
        await self.setPolicy(
            v8_types.EzspPolicyId.TRUST_CENTER_POLICY,
            self.tc_policy,
        )



import importlib.metadata

import pytest
import zigpy.exceptions
import zigpy.state
import zigpy.types as zigpy_t
import zigpy.zdo.types as zdo_t

from bellows.exception import EzspError
from bellows.ezsp import EZSP
from bellows.ezsp.v9.commands import GetTokenDataRsp
import bellows.types as t

from tests.async_mock import AsyncMock, PropertyMock
from tests.test_application import app, ezsp_mock, ieee, make_app


@pytest.fixture
def node_info():
    return zigpy.state.NodeInfo(
        nwk=zigpy_t.NWK(0x0000),
        ieee=zigpy_t.EUI64.convert("00:12:4b:00:1c:a1:b8:46"),
        logical_type=zdo_t.LogicalType.Coordinator,
        model="Mock board",
        manufacturer="Mock Manufacturer",
        version="Mock version",
    )


@pytest.fixture
def network_info(node_info):
    return zigpy.state.NetworkInfo(
        extended_pan_id=zigpy_t.ExtendedPanId.convert("bd:27:0b:38:37:95:dc:87"),
        pan_id=zigpy_t.PanId(0x9BB0),
        nwk_update_id=18,
        nwk_manager_id=zigpy_t.NWK(0x0000),
        channel=zigpy_t.uint8_t(15),
        channel_mask=zigpy_t.Channels.ALL_CHANNELS,
        security_level=zigpy_t.uint8_t(5),
        network_key=zigpy.state.Key(
            key=zigpy_t.KeyData.convert("2ccade06b3090c310315b3d574d3c85a"),
            seq=108,
            tx_counter=118785,
        ),
        tc_link_key=zigpy.state.Key(
            key=zigpy_t.KeyData(b"ZigBeeAlliance09"),
            partner_ieee=node_info.ieee,
            tx_counter=8712428,
        ),
        key_table=[
            zigpy.state.Key(
                key=zigpy_t.KeyData.convert(
                    "85:7C:05:00:3E:76:1A:F9:68:9A:49:41:6A:60:5C:76"
                ),
                tx_counter=3792973670,
                rx_counter=1083290572,
                seq=147,
                partner_ieee=zigpy_t.EUI64.convert("CC:CC:CC:FF:FE:E6:8E:CA"),
            ),
            zigpy.state.Key(
                key=zigpy_t.KeyData.convert(
                    "CA:02:E8:BB:75:7C:94:F8:93:39:D3:9C:B3:CD:A7:BE"
                ),
                tx_counter=2597245184,
                rx_counter=824424412,
                seq=19,
                partner_ieee=zigpy_t.EUI64.convert("EC:1B:BD:FF:FE:2F:41:A4"),
            ),
        ],
        children=[
            zigpy_t.EUI64.convert("00:0B:57:FF:FE:2B:D4:57"),
        ],
        # If exposed by the stack, NWK addresses of other connected devices on the network
        nwk_addresses={
            # Two routers
            zigpy_t.EUI64.convert("CC:CC:CC:FF:FE:E6:8E:CA"): zigpy_t.NWK(0x44CB),
            zigpy_t.EUI64.convert("EC:1B:BD:FF:FE:2F:41:A4"): zigpy_t.NWK(0x0702),
            # Child device
            zigpy_t.EUI64.convert("00:0b:57:ff:fe:2b:d4:57"): zigpy_t.NWK(0xC06B),
        },
        stack_specific={"ezsp": {"hashed_tclk": "abcdabcdabcdabcdabcdabcdabcdabcd"}},
        source=f"bellows@{importlib.metadata.version('bellows')}",
    )


def _mock_app_for_load(app, ezsp_ver=7):
    """Mock methods on the application and EZSP objects to run network state code."""
    ezsp = app._ezsp
    ezsp.ezsp_version = ezsp_ver
    type(ezsp).types = EZSP._BY_VERSION[ezsp_ver].types

    app._ensure_network_running = AsyncMock()
    ezsp.getNetworkParameters = AsyncMock(
        return_value=[
            t.EmberStatus.SUCCESS,
            t.EmberNodeType.COORDINATOR,
            t.EmberNetworkParameters(
                extendedPanId=t.ExtendedPanId.convert("bd:27:0b:38:37:95:dc:87"),
                panId=t.EmberPanId(0x9BB0),
                radioTxPower=8,
                radioChannel=15,
                joinMethod=t.EmberJoinMethod.USE_MAC_ASSOCIATION,
                nwkManagerId=t.EmberNodeId(0x0000),
                nwkUpdateId=18,
                channels=t.Channels.ALL_CHANNELS,
            ),
        ]
    )

    ezsp.getNodeId = AsyncMock(return_value=[t.EmberNodeId(0x0000)])
    ezsp.getEui64 = AsyncMock(return_value=[t.EUI64.convert("00:12:4b:00:1c:a1:b8:46")])

    def get_configuration_value(config_id):
        size = {
            app._ezsp.types.EzspConfigId.CONFIG_ADDRESS_TABLE_SIZE: t.uint8_t(20),
            app._ezsp.types.EzspConfigId.CONFIG_KEY_TABLE_SIZE: t.uint8_t(13),
            app._ezsp.types.EzspConfigId.CONFIG_SECURITY_LEVEL: t.uint8_t(5),
        }[config_id]

        return [app._ezsp.types.EmberStatus.SUCCESS, size]

    ezsp.getConfigurationValue = AsyncMock(side_effect=get_configuration_value)

    ezsp.getCurrentSecurityState = AsyncMock(
        return_value=(
            t.EmberStatus.SUCCESS,
            ezsp.types.EmberCurrentSecurityState(
                bitmask=(
                    t.EmberCurrentSecurityBitmask.TRUST_CENTER_USES_HASHED_LINK_KEY
                    | 64
                    | 32
                    | t.EmberCurrentSecurityBitmask.HAVE_TRUST_CENTER_LINK_KEY
                    | t.EmberCurrentSecurityBitmask.GLOBAL_LINK_KEY
                ),
                trustCenterLongAddress=t.EUI64.convert("00:12:4b:00:1c:a1:b8:46"),
            ),
        )
    )

    def get_key(key_type):
        key = {
            ezsp.types.EmberKeyType.CURRENT_NETWORK_KEY: ezsp.types.EmberKeyStruct(
                bitmask=(
                    t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                    | t.EmberKeyStructBitmask.KEY_HAS_SEQUENCE_NUMBER
                ),
                type=ezsp.types.EmberKeyType.CURRENT_NETWORK_KEY,
                key=t.KeyData.convert("2ccade06b3090c310315b3d574d3c85a"),
                outgoingFrameCounter=118785,
                incomingFrameCounter=0,
                sequenceNumber=108,
                partnerEUI64=t.EUI64.convert("00:00:00:00:00:00:00:00"),
            ),
            ezsp.types.EmberKeyType.TRUST_CENTER_LINK_KEY: ezsp.types.EmberKeyStruct(
                bitmask=(
                    t.EmberKeyStructBitmask.KEY_IS_AUTHORIZED
                    | t.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
                    | t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                ),
                type=ezsp.types.EmberKeyType.TRUST_CENTER_LINK_KEY,
                key=t.KeyData.convert("abcdabcdabcdabcdabcdabcdabcdabcd"),
                outgoingFrameCounter=8712428,
                incomingFrameCounter=0,
                sequenceNumber=0,
                partnerEUI64=t.EUI64.convert("00:12:4b:00:1c:a1:b8:46"),
            ),
        }[key_type]

        return (t.EmberStatus.SUCCESS, key)

    ezsp.getKey = AsyncMock(side_effect=get_key)

    if ezsp_ver >= 13:

        def export_key(security_context):
            key = {
                ezsp.types.sl_zb_sec_man_key_type_t.NETWORK: t.KeyData.convert(
                    "2ccade06b3090c310315b3d574d3c85a"
                ),
                ezsp.types.sl_zb_sec_man_key_type_t.TC_LINK: t.KeyData.convert(
                    "abcdabcdabcdabcdabcdabcdabcdabcd"
                ),
            }[security_context.core_key_type]

            return (key, t.EmberStatus.SUCCESS)

        ezsp.exportKey = AsyncMock(side_effect=export_key)
        ezsp.getNetworkKeyInfo = AsyncMock(
            return_value=[
                ezsp.types.sl_Status.SL_STATUS_OK,
                ezsp.types.sl_zb_sec_man_network_key_info_t(
                    network_key_set=True,
                    alternate_network_key_set=False,
                    network_key_sequence_number=108,
                    alt_network_key_sequence_number=0,
                    network_key_frame_counter=118785,
                ),
            ]
        )


async def test_load_network_info_no_devices(app, network_info, node_info):
    """Test `load_network_info(load_devices=False)`"""
    _mock_app_for_load(app)

    await app.load_network_info(load_devices=False)

    assert app.state.node_info == node_info
    assert app.state.network_info == network_info.replace(
        key_table=[],
        children=[],
        nwk_addresses={},
        metadata=app.state.network_info.metadata,
    )


async def test_load_network_info_no_key_set(app, network_info, node_info):
    """Test loading network info in v13+ when no network key is set."""
    _mock_app_for_load(app, ezsp_ver=13)

    app._ezsp.getNetworkKeyInfo = AsyncMock(
        return_value=[
            app._ezsp.types.sl_Status.SL_STATUS_OK,
            app._ezsp.types.sl_zb_sec_man_network_key_info_t(
                network_key_set=False,  # Not set
                alternate_network_key_set=False,
                network_key_sequence_number=108,
                alt_network_key_sequence_number=0,
                network_key_frame_counter=118785,
            ),
        ]
    )

    with pytest.raises(zigpy.exceptions.NetworkNotFormed):
        await app.load_network_info(load_devices=False)


@pytest.mark.parametrize("ezsp_ver", [4, 6, 7, 13])
async def test_load_network_info_with_devices(app, network_info, node_info, ezsp_ver):
    """Test `load_network_info(load_devices=True)`"""
    _mock_app_for_load(app, ezsp_ver)

    def get_child_data_v6(index):
        if index == 0:
            status = t.EmberStatus.SUCCESS
        else:
            status = t.EmberStatus.NOT_JOINED

        return (
            status,
            t.EmberNodeId(0xC06B),
            t.EUI64.convert("00:0b:57:ff:fe:2b:d4:57"),
            t.EmberNodeType.SLEEPY_END_DEVICE,
        )

    def get_child_data_v7(index):
        if index == 0:
            status = t.EmberStatus.SUCCESS
        else:
            status = t.EmberStatus.NOT_JOINED

        return (
            status,
            app._ezsp.types.EmberChildData(
                eui64=t.EUI64.convert("00:0b:57:ff:fe:2b:d4:57"),
                type=t.EmberNodeType.SLEEPY_END_DEVICE,
                id=t.EmberNodeId(0xC06B),
                phy=0,
                power=128,
                timeout=3,
            ),
        )

    app._ezsp.ezsp_version = ezsp_ver
    app._ezsp.getChildData = AsyncMock(
        side_effect={
            13: get_child_data_v7,
            7: get_child_data_v7,
            6: get_child_data_v6,
            4: get_child_data_v6,
        }[ezsp_ver]
    )

    if ezsp_ver < 13:

        def get_key_table_entry(index):
            if index == 0:
                return (
                    t.EmberStatus.SUCCESS,
                    app._ezsp.types.EmberKeyStruct(
                        bitmask=(
                            t.EmberKeyStructBitmask.KEY_IS_AUTHORIZED
                            | t.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
                            | t.EmberKeyStructBitmask.KEY_HAS_INCOMING_FRAME_COUNTER
                            | t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                        ),
                        type=app._ezsp.types.EmberKeyType.APPLICATION_LINK_KEY,
                        key=t.KeyData.convert("857C05003E761AF9689A49416A605C76"),
                        outgoingFrameCounter=3792973670,
                        incomingFrameCounter=1083290572,
                        sequenceNumber=147,
                        partnerEUI64=t.EUI64.convert("CC:CC:CC:FF:FE:E6:8E:CA"),
                    ),
                )
            elif index == 1:
                return (
                    t.EmberStatus.SUCCESS,
                    app._ezsp.types.EmberKeyStruct(
                        bitmask=(
                            t.EmberKeyStructBitmask.KEY_IS_AUTHORIZED
                            | t.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
                            | t.EmberKeyStructBitmask.KEY_HAS_INCOMING_FRAME_COUNTER
                            | t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                        ),
                        type=app._ezsp.types.EmberKeyType.APPLICATION_LINK_KEY,
                        key=t.KeyData.convert("CA02E8BB757C94F89339D39CB3CDA7BE"),
                        outgoingFrameCounter=2597245184,
                        incomingFrameCounter=824424412,
                        sequenceNumber=19,
                        partnerEUI64=t.EUI64.convert("EC:1B:BD:FF:FE:2F:41:A4"),
                    ),
                )
            elif index >= 12:
                status = t.EmberStatus.INDEX_OUT_OF_RANGE
            else:
                status = t.EmberStatus.TABLE_ENTRY_ERASED

            return (
                status,
                app._ezsp.types.EmberKeyStruct(
                    bitmask=t.EmberKeyStructBitmask(244),
                    type=app._ezsp.types.EmberKeyType(0x46),
                    key=t.KeyData.convert("b8a11c004b1200cdabcdabcdabcdabcd"),
                    outgoingFrameCounter=8192,
                    incomingFrameCounter=0,
                    sequenceNumber=0,
                    partnerEUI64=t.EUI64.convert("00:12:4b:00:1c:a1:b8:46"),
                ),
            )

        app._ezsp.getKeyTableEntry = AsyncMock(side_effect=get_key_table_entry)
    else:

        def export_link_key_by_index(index):
            if index == 0:
                return (
                    t.EUI64.convert("CC:CC:CC:FF:FE:E6:8E:CA"),
                    t.KeyData.convert("857C05003E761AF9689A49416A605C76"),
                    app._ezsp.types.sl_zb_sec_man_aps_key_metadata_t(
                        bitmask=(
                            t.EmberKeyStructBitmask.KEY_IS_AUTHORIZED
                            | t.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
                            | t.EmberKeyStructBitmask.KEY_HAS_INCOMING_FRAME_COUNTER
                            | t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                        ),
                        outgoing_frame_counter=3792973670,
                        incoming_frame_counter=1083290572,
                        ttl_in_seconds=0,
                    ),
                    app._ezsp.types.sl_Status.SL_STATUS_OK,
                )
            elif index == 1:
                return (
                    t.EUI64.convert("EC:1B:BD:FF:FE:2F:41:A4"),
                    t.KeyData.convert("CA02E8BB757C94F89339D39CB3CDA7BE"),
                    app._ezsp.types.sl_zb_sec_man_aps_key_metadata_t(
                        bitmask=(
                            t.EmberKeyStructBitmask.KEY_IS_AUTHORIZED
                            | t.EmberKeyStructBitmask.KEY_HAS_PARTNER_EUI64
                            | t.EmberKeyStructBitmask.KEY_HAS_INCOMING_FRAME_COUNTER
                            | t.EmberKeyStructBitmask.KEY_HAS_OUTGOING_FRAME_COUNTER
                        ),
                        outgoing_frame_counter=2597245184,
                        incoming_frame_counter=824424412,
                        ttl_in_seconds=0,
                    ),
                    app._ezsp.types.sl_Status.SL_STATUS_OK,
                )

            return (
                t.EUI64.convert("7f:c9:35:e1:b0:00:00:00"),
                t.KeyData.convert("80:45:38:73:55:00:00:00:08:e4:35:c9:7f:00:00:00"),
                app._ezsp.types.sl_zb_sec_man_aps_key_metadata_t(
                    bitmask=t.EmberKeyStructBitmask(43976),
                    outgoing_frame_counter=85,
                    incoming_frame_counter=0,
                    ttl_in_seconds=0,
                ),
                app._ezsp.types.sl_Status.SL_STATUS_NOT_FOUND,
            )

        app._ezsp.exportLinkKeyByIndex = AsyncMock(side_effect=export_link_key_by_index)

    def get_addr_table_node_id(index):
        return (
            {
                16: t.EmberNodeId(0x44CB),
                17: t.EmberNodeId(0x0702),
                18: t.EmberNodeId(0x0000),  # bogus entry
            }.get(index, t.EmberNodeId(0xFFFF)),
        )

    app._ezsp.getAddressTableRemoteNodeId = AsyncMock(
        side_effect=get_addr_table_node_id
    )

    def get_addr_table_eui64(index):
        if index < 16:
            return (t.EUI64.convert("ff:ff:ff:ff:ff:ff:ff:ff"),)
        elif 16 <= index <= 18:
            return (
                {
                    16: t.EUI64.convert("cc:cc:cc:ff:fe:e6:8e:ca"),
                    17: t.EUI64.convert("ec:1b:bd:ff:fe:2f:41:a4"),
                    18: t.EUI64.convert("00:00:00:00:00:00:00:00"),
                }[index],
            )
        else:
            return (t.EUI64.convert("00:00:00:00:00:00:00:00"),)

    app._ezsp.getAddressTableRemoteEui64 = AsyncMock(side_effect=get_addr_table_eui64)

    await app.load_network_info(load_devices=True)

    nwk_addresses = network_info.nwk_addresses

    # v4 doesn't support address table reads so the only known addresses are from the
    # `getChildData` call
    if ezsp_ver == 4:
        nwk_addresses = {
            ieee: addr
            for ieee, addr in network_info.nwk_addresses.items()
            if ieee in network_info.children
        }
    else:
        nwk_addresses = network_info.nwk_addresses

    # EZSP doesn't provide a command to set the key sequence number
    assert app.state.network_info == network_info.replace(
        key_table=[key.replace(seq=0) for key in network_info.key_table],
        nwk_addresses=nwk_addresses,
        metadata=app.state.network_info.metadata,
        # TC link key does not have a TX counter
        tc_link_key=network_info.tc_link_key.replace(
            tx_counter=app.state.network_info.tc_link_key.tx_counter
        ),
    )
    assert app.state.node_info == node_info

    # No crash-prone calls were made
    if ezsp_ver == 4:
        app._ezsp.getAddressTableRemoteNodeId.assert_not_called()
        app._ezsp.getAddressTableRemoteEui64.assert_not_called()


def _mock_app_for_write(app, network_info, node_info, ezsp_ver=7):
    ezsp = app._ezsp
    ezsp.ezsp_version = ezsp_ver
    type(ezsp).types = EZSP._BY_VERSION[ezsp_ver].types

    network_state = ezsp.types.EmberNetworkStatus.JOINED_NETWORK
    ezsp.networkState = AsyncMock(side_effect=lambda: [network_state])

    def leave_network():
        nonlocal network_state
        network_state = ezsp.types.EmberNetworkStatus.NO_NETWORK

        return [t.EmberStatus.NETWORK_DOWN]

    def form_network(params):
        nonlocal network_state
        network_state = ezsp.types.EmberNetworkStatus.JOINED_NETWORK

        return [t.EmberStatus.SUCCESS]

    ezsp.leaveNetwork = AsyncMock(side_effect=leave_network)
    ezsp.formNetwork = AsyncMock(side_effect=form_network)

    ezsp.getEui64 = AsyncMock(return_value=[t.EUI64.convert("00:12:4b:00:1c:a1:b8:46")])

    ezsp.setInitialSecurityState = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp.clearKeyTable = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp.getConfigurationValue = AsyncMock(
        return_value=[t.EmberStatus.SUCCESS, t.uint8_t(200)]
    )

    if ezsp_ver >= 13:
        ezsp.importLinkKey = AsyncMock(
            side_effect=[
                # Only the first one succeeds
                (t.EmberStatus.SUCCESS,),
            ]
            + [
                # The rest will fail
                (t.EmberStatus.TABLE_FULL,),
            ]
            * 20
        )
    else:
        ezsp.addOrUpdateKeyTableEntry = AsyncMock(
            side_effect=[
                # Only the first one succeeds
                (t.EmberStatus.SUCCESS,),
            ]
            + [
                # The rest will fail
                (t.EmberStatus.TABLE_FULL,),
            ]
            * 20
        )

    if ezsp_ver == 4:
        ezsp.setValue = AsyncMock(return_value=[t.EmberStatus.BAD_ARGUMENT])
    else:
        ezsp.setValue = AsyncMock(return_value=[t.EmberStatus.SUCCESS])

    if ezsp_ver >= 9:
        ezsp.setChildData = AsyncMock(return_value=[t.EmberStatus.SUCCESS])

    ezsp.setValue = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp.setMfgToken = AsyncMock(return_value=[t.EmberStatus.SUCCESS])
    ezsp.getTokenData = AsyncMock(
        return_value=GetTokenDataRsp(status=t.EmberStatus.LIBRARY_NOT_PRESENT)
    )


@pytest.mark.parametrize("ezsp_ver", [4, 7, 13])
async def test_write_network_info(app, network_info, node_info, ezsp_ver):
    _mock_app_for_write(app, network_info, node_info, ezsp_ver)

    await app.write_network_info(
        network_info=network_info.replace(
            children=network_info.children
            + [
                # Bogus child that can't be restored
                zigpy_t.EUI64.convert("FF:FF:57:FF:FE:2B:D4:57")
            ]
        ),
        node_info=node_info,
    )


@pytest.mark.parametrize("can_burn", [True, False])
@pytest.mark.parametrize("can_rewrite", [True, False])
@pytest.mark.parametrize("force_write", [True, False])
async def test_write_network_info_write_eui64(
    caplog, app, network_info, node_info, can_burn, can_rewrite, force_write
):
    _mock_app_for_write(app, network_info, node_info)

    # Differs from what is in `node_info`
    app._ezsp.getEui64.return_value = [t.EUI64.convert("AA:AA:AA:AA:AA:AA:AA:AA")]
    app._ezsp.can_burn_userdata_custom_eui64 = AsyncMock(return_value=can_burn)
    app._ezsp.can_rewrite_custom_eui64 = AsyncMock(return_value=can_rewrite)

    await app.write_network_info(
        network_info=network_info.replace(
            stack_specific={
                "ezsp": {
                    "i_understand_i_can_update_eui64_only_once_and_i_still_want_to_do_it": force_write,
                    **network_info.stack_specific["ezsp"],
                }
            }
        ),
        node_info=node_info,
    )

    if can_rewrite:
        # Everything else is always ignored
        app._ezsp.write_custom_eui64.assert_called_once_with(node_info.ieee)
    elif can_burn and force_write:
        # Can only be forced
        app._ezsp.write_custom_eui64.assert_called_once_with(
            node_info.ieee, burn_into_userdata=True
        )
    else:
        # It is not written otherwise
        app._ezsp.write_custom_eui64.assert_not_called()

        if can_burn and not force_write:
            assert "does not match" in caplog.text
        elif not can_burn and force_write:
            assert "has already been written" in caplog.text
        elif not can_burn and not force_write:
            assert "does not match" in caplog.text
        else:
            assert False


async def test_write_network_info_generate_hashed_tclk(app, network_info, node_info):
    _mock_app_for_write(app, network_info, node_info)

    seen_keys = set()

    for i in range(10):
        app._ezsp.setInitialSecurityState.reset_mock()

        await app.write_network_info(
            network_info=network_info.replace(stack_specific={}),
            node_info=node_info,
        )

        call = app._ezsp.setInitialSecurityState.mock_calls[0]
        seen_keys.add(tuple(call[1][0].preconfiguredKey))

    # A new hashed key is randomly generated each time if none is provided
    assert len(seen_keys) == 10


async def test_reset_network_with_no_formed_network(app):
    _mock_app_for_write(app, network_info, node_info)

    app._ezsp.networkState = AsyncMock(
        return_value=[app._ezsp.types.EmberNetworkStatus.NO_NETWORK]
    )

    app._ezsp.networkInit = AsyncMock(return_value=[t.EmberStatus.NOT_JOINED])

    await app.reset_network_info()



import asyncio
import logging
import sys

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

import zigpy.config
import zigpy.serial

from bellows.ash import AshProtocol
from bellows.thread import EventLoopThread, ThreadsafeProxy
import bellows.types as t

LOGGER = logging.getLogger(__name__)
RESET_TIMEOUT = 5


class Gateway(asyncio.Protocol):
    FLAG = b"\x7E"  # Marks end of frame
    ESCAPE = b"\x7D"
    XON = b"\x11"  # Resume transmission
    XOFF = b"\x13"  # Stop transmission
    SUBSTITUTE = b"\x18"
    CANCEL = b"\x1A"  # Terminates a frame in progress
    STUFF = 0x20
    RANDOMIZE_START = 0x42
    RANDOMIZE_SEQ = 0xB8

    RESERVED = FLAG + ESCAPE + XON + XOFF + SUBSTITUTE + CANCEL

    class Terminator:
        pass

    def __init__(self, application, connected_future=None, connection_done_future=None):
        self._application = application

        self._reset_future = None
        self._startup_reset_future = None
        self._connected_future = connected_future
        self._connection_done_future = connection_done_future

        self._transport = None

    def close(self):
        self._transport.close()

    def connection_made(self, transport):
        """Callback when the uart is connected"""
        self._transport = transport
        if self._connected_future is not None:
            self._connected_future.set_result(True)

    async def send_data(self, data: bytes) -> None:
        await self._transport.send_data(data)

    def data_received(self, data):
        """Callback when there is data received from the uart"""
        self._application.frame_received(data)

    def reset_received(self, code: t.NcpResetCode) -> None:
        """Reset acknowledgement frame receive handler"""
        # not a reset we've requested. Signal application reset
        if code is not t.NcpResetCode.RESET_SOFTWARE:
            self._application.enter_failed_state(code)
            return

        if self._reset_future and not self._reset_future.done():
            self._reset_future.set_result(True)
        elif self._startup_reset_future and not self._startup_reset_future.done():
            self._startup_reset_future.set_result(True)
        else:
            LOGGER.warning("Received an unexpected reset: %r", code)

    def error_received(self, code: t.NcpResetCode) -> None:
        """Error frame receive handler."""
        self._application.enter_failed_state(code)

    async def wait_for_startup_reset(self) -> None:
        """Wait for the first reset frame on startup."""
        assert self._startup_reset_future is None
        self._startup_reset_future = asyncio.get_running_loop().create_future()

        try:
            await self._startup_reset_future
        finally:
            self._startup_reset_future = None

    def _reset_cleanup(self, future):
        """Delete reset future."""
        self._reset_future = None

    def eof_received(self):
        """Server gracefully closed its side of the connection."""
        self.connection_lost(ConnectionResetError("Remote server closed connection"))

    def connection_lost(self, exc):
        """Port was closed unexpectedly."""

        LOGGER.debug("Connection lost: %r", exc)
        reason = exc or ConnectionResetError("Remote server closed connection")

        # XXX: The startup reset future must be resolved with an error *before* the
        # "connection done" future is completed: the secondary thread has an attached
        # callback to stop itself, which will cause the a future to propagate a
        # `CancelledError` into the active event loop, breaking everything!
        if self._startup_reset_future:
            self._startup_reset_future.set_exception(reason)

        if self._connection_done_future:
            self._connection_done_future.set_result(exc)
            self._connection_done_future = None

        if self._reset_future:
            self._reset_future.set_exception(reason)
            self._reset_future = None

        if exc is None:
            LOGGER.debug("Closed serial connection")
            return

        LOGGER.error("Lost serial connection: %r", exc)
        self._application.connection_lost(exc)

    async def reset(self):
        """Send a reset frame and init internal state."""
        LOGGER.debug("Resetting ASH")
        if self._reset_future is not None:
            LOGGER.error(
                "received new reset request while an existing one is in progress"
            )
            return await self._reset_future

        self._transport.send_reset()
        self._reset_future = asyncio.get_event_loop().create_future()
        self._reset_future.add_done_callback(self._reset_cleanup)

        async with asyncio_timeout(RESET_TIMEOUT):
            return await self._reset_future


async def _connect(config, application):
    loop = asyncio.get_event_loop()

    connection_future = loop.create_future()
    connection_done_future = loop.create_future()

    gateway = Gateway(application, connection_future, connection_done_future)
    protocol = AshProtocol(gateway)

    if config[zigpy.config.CONF_DEVICE_FLOW_CONTROL] is None:
        xon_xoff, rtscts = True, False
    else:
        xon_xoff, rtscts = False, True

    transport, _ = await zigpy.serial.create_serial_connection(
        loop,
        lambda: protocol,
        url=config[zigpy.config.CONF_DEVICE_PATH],
        baudrate=config[zigpy.config.CONF_DEVICE_BAUDRATE],
        xonxoff=xon_xoff,
        rtscts=rtscts,
    )

    await connection_future

    thread_safe_protocol = ThreadsafeProxy(gateway, loop)
    return thread_safe_protocol, connection_done_future


async def connect(config, application, use_thread=True):
    if use_thread:
        application = ThreadsafeProxy(application, asyncio.get_event_loop())
        thread = EventLoopThread()
        await thread.start()
        try:
            protocol, connection_done = await thread.run_coroutine_threadsafe(
                _connect(config, application)
            )
        except Exception:
            thread.force_stop()
            raise
        connection_done.add_done_callback(lambda _: thread.force_stop())
    else:
        protocol, _ = await _connect(config, application)
    return protocol



import logging

import click
import click_log

from . import opts


@click.group()
@click_log.simple_verbosity_option(logging.getLogger(), default="WARNING")
@opts.device
@opts.baudrate
@opts.flow_control
@click.pass_context
def main(ctx, device, baudrate, flow_control):
    ctx.obj = {
        "device": device,
        "baudrate": baudrate,
        "flow_control": flow_control,
    }
    click_log.basic_config()


if __name__ == "__main__":
    main()



from zigpy.types import Struct as EzspStruct, StructField

from . import basic, named


class EmberNetworkParameters(EzspStruct):
    # Network parameters.
    # The network's extended PAN identifier.
    extendedPanId: named.ExtendedPanId
    # The network's PAN identifier.
    panId: named.EmberPanId
    # A power setting, in dBm.
    radioTxPower: basic.uint8_t
    # A radio channel.
    radioChannel: basic.uint8_t
    # The method used to initially join the network.
    joinMethod: named.EmberJoinMethod
    # NWK Manager ID. The ID of the network manager in the current network.
    # This may only be set at joining when using USE_NWK_COMMISSIONING as
    # the join method.
    nwkManagerId: named.EmberNodeId
    # NWK Update ID. The value of the ZigBee nwkUpdateId known by the
    # stack. This is used to determine the newest instance of the network
    # after a PAN ID or channel change. This may only be set at joining
    # when using USE_NWK_COMMISSIONING as the join method.
    nwkUpdateId: basic.uint8_t
    # NWK channel mask. The list of preferred channels that the NWK manager
    # has told this device to use when searching for the network. This may
    # only be set at joining when using USE_NWK_COMMISSIONING as the join
    # method.
    channels: named.Channels


class EmberZigbeeNetwork(EzspStruct):
    # The parameters of a ZigBee network.
    # The 802.15.4 channel associated with the network.
    channel: basic.uint8_t
    # The network's PAN identifier.
    panId: named.EmberPanId
    # The network's extended PAN identifier.
    extendedPanId: named.ExtendedPanId
    # Whether the network is allowing MAC associations.
    allowingJoin: named.Bool
    # The Stack Profile associated with the network.
    stackProfile: basic.uint8_t
    # The instance of the Network.
    nwkUpdateId: basic.uint8_t


class EmberApsFrame(EzspStruct):
    # ZigBee APS frame parameters.
    # The application profile ID that describes the format of the message.
    profileId: basic.uint16_t
    # The cluster ID for this message.
    clusterId: basic.uint16_t
    # The source endpoint.
    sourceEndpoint: basic.uint8_t
    # The destination endpoint.
    destinationEndpoint: basic.uint8_t
    # A bitmask of options.
    options: named.EmberApsOption
    # The group ID for this message, if it is multicast mode.
    groupId: basic.uint16_t
    # The sequence number.
    sequence: basic.uint8_t


class EmberBindingTableEntry(EzspStruct):
    # An entry in the binding table.
    # The type of binding.
    type: named.EmberBindingType
    # The endpoint on the local node.
    local: basic.uint8_t
    # A cluster ID that matches one from the local endpoint's simple
    # descriptor. This cluster ID is set by the provisioning application to
    # indicate which part an endpoint's functionality is bound to this
    # particular remote node and is used to distinguish between unicast and
    # multicast bindings. Note that a binding can be used to send messages
    # with any cluster ID, not just that listed in the binding.
    clusterId: basic.uint16_t
    # The endpoint on the remote node (specified by identifier).
    remote: basic.uint8_t
    # A 64-bit identifier. This is either the destination EUI64 (for
    # unicasts) or the 64-bit group address (for multicasts).
    identifier: named.EUI64
    # The index of the network the binding belongs to.
    networkIndex: basic.uint8_t


class EmberMulticastTableEntry(EzspStruct):
    # A multicast table entry indicates that a particular endpoint is a member
    # of a particular multicast group.  Only devices with an endpoint in a
    # multicast group will receive messages sent to that multicast group.
    # The multicast group ID.
    multicastId: named.EmberMulticastId
    # The endpoint that is a member, or 0 if this entry is not in use (the
    # ZDO is not a member of any multicast groups.)
    endpoint: basic.uint8_t
    # The network index of the network the entry is related to.
    networkIndex: basic.uint8_t = StructField(optional=True)


class EmberTransientKeyData(EzspStruct):
    # The transient key data structure. Added in ver. 5
    # The IEEE address paired with the transient link key.
    eui64: named.EUI64
    # The key data structure matching the transient key.
    keyData: named.KeyData
    # The incoming frame counter associated with this key.
    incomingFrameCounter: basic.uint32_t
    # The number of milliseconds remaining before the key
    # is automatically timed out of the transient key table.
    countdownTimerMs: basic.uint32_t


class EmberAesMmoHashContext(EzspStruct):
    # The hash context for an ongoing hash operation.
    # The result of ongoing the hash operation.
    result: basic.FixedList[basic.uint8_t, 16]
    # The total length of the data that has been hashed so far.
    length: basic.uint32_t


class EmberNeighborTableEntry(EzspStruct):
    # A neighbor table entry stores information about the reliability of RF
    # links to and from neighboring nodes.
    # The neighbor's two byte network id
    shortId: named.EmberNodeId
    # An exponentially weighted moving average of the link quality values
    # of incoming packets from this neighbor as reported by the PHY.
    averageLqi: basic.uint8_t
    # The incoming cost for this neighbor, computed from the average LQI.
    # Values range from 1 for a good link to 7 for a bad link.
    inCost: basic.uint8_t
    # The outgoing cost for this neighbor, obtained from the most recently
    # received neighbor exchange message from the neighbor. A value of zero
    # means that a neighbor exchange message from the neighbor has not been
    # received recently enough, or that our id was not present in the most
    # recently received one.
    outCost: basic.uint8_t
    # The number of aging periods elapsed since a link status message was
    # last received from this neighbor. The aging period is 16 seconds.
    age: basic.uint8_t
    # The 8 byte EUI64 of the neighbor.
    longId: named.EUI64


class EmberRouteTableEntry(EzspStruct):
    # A route table entry stores information about the next hop along the route
    # to the destination.
    # The short id of the destination. A value of 0xFFFF indicates the
    # entry is unused.
    destination: named.EmberNodeId
    # The short id of the next hop to this destination.
    nextHop: basic.uint16_t
    # Indicates whether this entry is active (0), being discovered (1)),
    # unused (3), or validating (4).
    status: basic.uint8_t
    # The number of seconds since this route entry was last used to send a
    # packet.
    age: basic.uint8_t
    # Indicates whether this destination is a High RAM Concentrator (2), a
    # Low RAM Concentrator (1), or not a concentrator (0).
    concentratorType: basic.uint8_t
    # For a High RAM Concentrator, indicates whether a route record is
    # needed (2), has been sent (1), or is no long needed (0) because a
    # source routed message from the concentrator has been received.
    routeRecordState: basic.uint8_t


class EmberInitialSecurityState(EzspStruct):
    # The security data used to set the configuration for the stack, or the
    # retrieved configuration currently in use.
    # A bitmask indicating the security state used to indicate what the
    # security configuration will be when the device forms or joins the
    # network.
    bitmask: named.EmberInitialSecurityBitmask
    # The pre-configured Key data that should be used when forming or
    # joining the network. The security bitmask must be set with the
    # HAVE_PRECONFIGURED_KEY bit to indicate that the key contains valid
    # data.
    preconfiguredKey: named.KeyData
    # The Network Key that should be used by the Trust Center when it forms
    # the network, or the Network Key currently in use by a joined device.
    # The security bitmask must be set with HAVE_NETWORK_KEY to indicate
    # that the key contains valid data.
    networkKey: named.KeyData
    # The sequence number associated with the network key. This is only
    # valid if the HAVE_NETWORK_KEY has been set in the security bitmask.
    networkKeySequenceNumber: basic.uint8_t
    # This is the long address of the trust center on the network that will
    # be joined. It is usually NOT set prior to joining the network and
    # instead it is learned during the joining message exchange. This field
    # is only examined if HAVE_TRUST_CENTER_EUI64 is set in the
    # EmberInitialSecurityState::bitmask. Most devices should clear that
    # bit and leave this field alone. This field must be set when using
    # commissioning mode.
    preconfiguredTrustCenterEui64: named.EUI64


class EmberCurrentSecurityState(EzspStruct):
    # The security options and information currently used by the stack.
    # A bitmask indicating the security options currently in use by a
    # device joined in the network.
    bitmask: named.EmberCurrentSecurityBitmask
    # The IEEE Address of the Trust Center device.
    trustCenterLongAddress: named.EUI64


class EmberZllSecurityAlgorithmData(EzspStruct):
    # Data associated with the ZLL security algorithm.
    # Transaction identifier.
    transactionId: basic.uint32_t
    # Response identifier.
    responseId: basic.uint32_t
    # Bitmask.
    bitmask: basic.uint16_t


class EmberZllNetwork(EzspStruct):
    # The parameters of a ZLL network.
    # The parameters of a ZigBee network.
    zigbeeNetwork: EmberZigbeeNetwork
    # Data associated with the ZLL security algorithm.
    securityAlgorithm: EmberZllSecurityAlgorithmData
    # Associated EUI64.
    eui64: named.EUI64
    # The node id.
    nodeId: named.EmberNodeId
    # The ZLL state.
    state: named.EmberZllState
    # The node type.
    nodeType: named.EmberNodeType
    # The number of sub devices.
    numberSubDevices: basic.uint8_t
    # The total number of group identifiers.
    totalGroupIdentifiers: basic.uint8_t
    # RSSI correction value.
    rssiCorrection: basic.uint8_t


class EmberZllInitialSecurityState(EzspStruct):
    # Describes the initial security features and requirements that will be
    # used when forming or joining ZLL networks.
    # Unused bitmask; reserved for future use.
    bitmask: basic.uint32_t
    # The key encryption algorithm advertised by the application.
    keyIndex: named.EmberZllKeyIndex
    # The encryption key for use by algorithms that require it.
    encryptionKey: named.KeyData
    # The pre-configured link key used during classical ZigBee
    # commissioning.
    preconfiguredKey: named.KeyData


class EmberZllDeviceInfoRecord(EzspStruct):
    # Information about a specific ZLL Device.
    # EUI64 associated with the device.
    ieeeAddress: named.EUI64
    # Endpoint id.
    endpointId: basic.uint8_t
    # Profile id.
    profileId: basic.uint16_t
    # Device id.
    deviceId: basic.uint16_t
    # Associated version.
    version: basic.uint8_t
    # Number of relevant group ids.
    groupIdCount: basic.uint8_t


class EmberZllAddressAssignment(EzspStruct):
    # ZLL address assignment data.
    # Relevant node id.
    nodeId: named.EmberNodeId
    # Minimum free node id.
    freeNodeIdMin: named.EmberNodeId
    # Maximum free node id.
    freeNodeIdMax: named.EmberNodeId
    # Minimum group id.
    groupIdMin: named.EmberMulticastId
    # Maximum group id.
    groupIdMax: named.EmberMulticastId
    # Minimum free group id.
    freeGroupIdMin: named.EmberMulticastId
    # Maximum free group id.
    freeGroupIdMax: named.EmberMulticastId


class EmberTokenData(EzspStruct):
    # Token Data
    # Token data size in bytes
    size: basic.uint32_t
    # Token data pointer
    data: basic.uint8_t


class EmberTokenInfo(EzspStruct):
    # Information of a token in the token table
    # NVM3 key of the token
    nvm3Key: named.NV3KeyId
    # Token is a counter type
    isCnt: named.Bool
    # Token is an indexed token
    isIdx: named.Bool
    # Size of the token
    size: basic.uint8_t
    # Array size of the token
    arraySize: basic.uint8_t


class EmberTokTypeStackZllData(EzspStruct):
    # Public API for ZLL stack data token.
    # Token bitmask.
    bitmask: basic.uint32_t
    # Minimum free node id.
    freeNodeIdMin: basic.uint16_t
    # Maximum free node id.
    freeNodeIdMax: basic.uint16_t
    # Local minimum group id.
    myGroupIdMin: basic.uint16_t
    # Minimum free group id.
    freeGroupIdMin: basic.uint16_t
    # Maximum free group id.
    freeGroupIdMax: basic.uint16_t
    # RSSI correction value.
    rssiCorrection: basic.uint8_t


class EmberTokTypeStackZllSecurity(EzspStruct):
    # Public API for ZLL stack security token.
    # Token bitmask.
    bitmask: basic.uint32_t
    # Key index.
    keyIndex: basic.uint8_t
    # Encryption key.
    encryptionKey: named.KeyData
    # Preconfigured key.
    preconfiguredKey: named.KeyData


class EmberGpAddress(EzspStruct):
    # A GP address structure.
    # The GPD's EUI64.
    gpdIeeeAddress: named.EUI64
    # The GPD's source ID.
    sourceId: basic.uint32_t
    # The GPD Application ID.
    applicationId: basic.uint8_t
    # The GPD endpoint.
    endpoint: basic.uint8_t


class NV3StackTrustCenterToken(EzspStruct):
    """NV3 stack trust center token value."""

    mode: basic.uint16_t
    eui64: named.EUI64
    key: named.KeyData



import pytest
import zigpy.zdo.types as zdo_t

import bellows.types as t


def test_lvbytes32():
    d, r = t.LVBytes32.deserialize(b"\x04\x00\x00\x0012345")
    assert r == b"5"
    assert d == b"1234"

    assert t.LVBytes32.serialize(d) == b"\x04\x00\x00\x001234"


@pytest.mark.parametrize(
    "node_type, logical_type",
    (
        (0, 7),
        (1, 0),
        (2, 1),
        (3, 2),
        (4, 7),
        (0xFF, 7),
    ),
)
def test_ember_node_type_to_zdo_logical_type(node_type, logical_type):
    """Test conversion of node type to logical type."""

    node_type = t.EmberNodeType(node_type)
    assert node_type.zdo_logical_type == zdo_t.LogicalType(logical_type)



import voluptuous as vol

from bellows.config import cv_optional_int, cv_uint16

from .types import EzspConfigId, EzspDecisionId, EzspPolicyId

EZSP_SCHEMA = {
    #
    # The number of packet buffers available to the stack. When set to the special
    # value 0xFF, the NCP will allocate all remaining configuration RAM towards
    # packet buffers, such that the resulting count will be the largest whole number
    # of packet buffers that can fit into the available memory
    vol.Optional(EzspConfigId.CONFIG_PACKET_BUFFER_COUNT.name): cv_optional_int(
        min=1, max=255
    ),
    # The maximum number of router neighbors the stack can keep track of. A neighbor
    # is a node within radio range
    vol.Optional(EzspConfigId.CONFIG_NEIGHBOR_TABLE_SIZE.name): cv_optional_int(
        min=8, max=16
    ),
    #
    # The maximum number of APS retried messages the stack can be transmitting at
    # any time
    vol.Optional(EzspConfigId.CONFIG_APS_UNICAST_MESSAGE_COUNT.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The maximum number of non-volatile bindings supported by the stack
    vol.Optional(EzspConfigId.CONFIG_BINDING_TABLE_SIZE.name): cv_optional_int(
        min=0, max=32
    ),
    #
    # The maximum number of EUI64 to network address associations that the stack can
    # maintain for the application. (Note: The total number of such address
    # associations maintained by the NCP is the sum of the value of this setting and
    # the value of EZSP_CONFIG_TRUST_CENTER_ADDRESS_CA CHE_SIZE)
    vol.Optional(EzspConfigId.CONFIG_ADDRESS_TABLE_SIZE.name): cv_optional_int(min=0),
    #
    # The maximum number of multicast groups that the device may be a member of
    vol.Optional(EzspConfigId.CONFIG_MULTICAST_TABLE_SIZE.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The maximum number of destinations to which a node can route messages. This
    # includes both messages originating at this node and those relayed for others
    vol.Optional(EzspConfigId.CONFIG_ROUTE_TABLE_SIZE.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The number of simultaneous route discoveries that a node will support
    vol.Optional(EzspConfigId.CONFIG_DISCOVERY_TABLE_SIZE.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The size of the alarm broadcast buffer
    vol.Optional(EzspConfigId.CONFIG_BROADCAST_ALARM_DATA_SIZE.name): cv_optional_int(
        min=0, max=16
    ),
    #
    # The size of the unicast alarm buffers allocated for end device children
    vol.Optional(EzspConfigId.CONFIG_UNICAST_ALARM_DATA_SIZE.name): cv_optional_int(
        min=0, max=16
    ),
    #
    # Specifies the stack profile
    vol.Optional(EzspConfigId.CONFIG_STACK_PROFILE.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The security level used for security at the MAC and network layers. The
    # supported values are 0 (no security) and 5 (payload is encrypted and a
    # four-byte MIC is used for authentication)
    vol.Optional(EzspConfigId.CONFIG_SECURITY_LEVEL.name): cv_optional_int(
        min=0, max=5
    ),
    #
    # The maximum number of hops for a message
    vol.Optional(EzspConfigId.CONFIG_MAX_HOPS.name): cv_optional_int(min=0, max=30),
    #
    # The maximum number of end device children that a router will support
    vol.Optional(EzspConfigId.CONFIG_MAX_END_DEVICE_CHILDREN.name): cv_optional_int(
        min=0, max=32
    ),
    #
    # The maximum amount of time that the MAC will hold a message for indirect
    # transmission to a child
    vol.Optional(
        EzspConfigId.CONFIG_INDIRECT_TRANSMISSION_TIMEOUT.name
    ): cv_optional_int(min=0, max=30000),
    #
    # The maximum amount of time that an end device child can wait between polls.
    # If no poll is heard within this timeout, then the parent removes the end
    # device from its tables. The timeout corresponding to a value of zero is 10
    # seconds. The timeout corresponding to a nonzero value N is 2^N minutes,
    # ranging from 2^1 = 2 minutes to 2^14 = 16384 minutes
    vol.Optional(EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The maximum amount of time that a mobile node can wait between polls. If no
    # poll is heard within this timeout, then the parent removes the mobile node
    # from its tables
    vol.Optional(EzspConfigId.CONFIG_MOBILE_NODE_POLL_TIMEOUT.name): cv_optional_int(
        min=0
    ),
    #
    # The number of child table entries reserved for use only by mobile nodes
    vol.Optional(
        EzspConfigId.CONFIG_RESERVED_MOBILE_CHILD_ENTRIES.name
    ): cv_optional_int(min=0, max=254),
    #
    # Enables boost power mode and/or the alternate transmitter output
    vol.Optional(EzspConfigId.CONFIG_TX_POWER_MODE.name): cv_optional_int(min=0, max=3),
    #
    # 0: Allow this node to relay messages. 1: Prevent this node from relaying
    # messages
    vol.Optional(EzspConfigId.CONFIG_DISABLE_RELAY.name): cv_optional_int(min=0, max=1),
    #
    # The maximum number of EUI64 to network address associations that the Trust
    # Center can maintain. These address cache entries are reserved for and reused
    # by the Trust Center when processing device join/rejoin authentications. This
    # cache size limits the number of overlapping joins the Trust Center can process
    # within a narrow time window (e.g. two seconds), and thus should be set to the
    # maximum number of near simultaneous joins the Trust Center is expected to
    # accommodate. (Note: The total number of such address associations maintained
    # by the NCP is the sum of the value of this setting and the value of
    # EZSP_CONFIG_ADDRESS_TABLE_SIZE)
    vol.Optional(
        EzspConfigId.CONFIG_TRUST_CENTER_ADDRESS_CACHE_SIZE.name
    ): cv_optional_int(min=0),
    #
    # The size of the source route table
    vol.Optional(EzspConfigId.CONFIG_SOURCE_ROUTE_TABLE_SIZE.name): cv_optional_int(
        min=0
    ),
    # The units used for timing out end devices on their parent
    vol.Optional(
        EzspConfigId.CONFIG_END_DEVICE_POLL_TIMEOUT_SHIFT.name
    ): cv_optional_int(min=0, max=10),
    #
    # The number of blocks of a fragmented message that can be sent in a single
    # window
    vol.Optional(EzspConfigId.CONFIG_FRAGMENT_WINDOW_SIZE.name): cv_optional_int(
        min=0, max=8
    ),
    #
    # The time (ms) the stack will wait between sending blocks of a fragmented
    # message
    vol.Optional(EzspConfigId.CONFIG_FRAGMENT_DELAY_MS.name): cv_optional_int(min=0),
    #
    # The size of the Key Table used for storing individual link keys (if the device
    # is a Trust Center) or Application Link Keys (if the device is a normal node)
    vol.Optional(EzspConfigId.CONFIG_KEY_TABLE_SIZE.name): cv_optional_int(min=0),
    #
    # The APS ACK timeout value (ms). The stack waits this amount of time between
    # resends of APS retried messages
    vol.Optional(EzspConfigId.CONFIG_APS_ACK_TIMEOUT.name): cv_optional_int(min=0),
    #
    # The duration of an active scan 15/4 scan duration units. This also controls
    # the jitter used when responding to a beacon request
    vol.Optional(EzspConfigId.CONFIG_ACTIVE_SCAN_DURATION.name): cv_optional_int(
        min=0, max=6
    ),
    #
    # The time (seoonds) the coordinator will wait for a second end device bind
    # request to arrive
    vol.Optional(EzspConfigId.CONFIG_END_DEVICE_BIND_TIMEOUT.name): cv_optional_int(
        min=1
    ),
    #
    # The number of PAN id conflict reports that must be received by the network
    # manager within one minute to trigger a PAN id change
    vol.Optional(
        EzspConfigId.CONFIG_PAN_ID_CONFLICT_REPORT_THRESHOLD.name
    ): cv_optional_int(min=1, max=63),
    #
    # The timeout value in minutes for how long the Trust Center or a normal node
    # waits for the ZigBee Request Key to complete. On the Trust Center this
    # controls whether or not the device buffers the request, waiting for a matching
    # pair of ZigBee Request Key. If the value is non-zero, the Trust Center buffers
    # and waits for that amount of time. If the value is zero, the Trust Center does
    # not buffer the request and immediately responds to the request. Zero is the
    # most compliant behavior
    vol.Optional(EzspConfigId.CONFIG_REQUEST_KEY_TIMEOUT.name): cv_optional_int(
        min=0, max=10
    ),
    #
    # This value indicates the size of the runtime modifiable certificate table.
    # Normally certificates are stored in MFG tokens but this table can be used to
    # field upgrade devices with new Smart Energy certificates. This value cannot be
    # set, it can only be queried
    vol.Optional(EzspConfigId.CONFIG_CERTIFICATE_TABLE_SIZE.name): cv_optional_int(
        min=0, max=1
    ),
    #
    # This is a bitmask that controls which incoming ZDO request messages are passed
    # to the application. The bits are defined in the EmberZdoConfigurationFlags
    # enumeration. To see if the application is required to send a ZDO response in
    # reply to an incoming message, the application must check the APS options
    # bitfield within the incomingMessageHandler callback to see if the
    # EMBER_APS_OPTION_ZDO_RESPONSE_REQUIRED flag is set
    vol.Optional(EzspConfigId.CONFIG_APPLICATION_ZDO_FLAGS.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # The maximum number of broadcasts during a single broadcast timeout period
    vol.Optional(EzspConfigId.CONFIG_BROADCAST_TABLE_SIZE.name): cv_optional_int(
        min=15, max=254
    ),
    #
    # The size of the MAC filter list table
    vol.Optional(EzspConfigId.CONFIG_MAC_FILTER_TABLE_SIZE.name): cv_optional_int(
        min=0, max=254
    ),
    #
    # The number of supported networks
    vol.Optional(EzspConfigId.CONFIG_SUPPORTED_NETWORKS.name): cv_optional_int(
        min=1, max=2
    ),
    #
    # Whether multicasts are sent to the RxOnWhenIdle=true address (0xFFFD) or the
    # sleepy broadcast address (0xFFFF). The RxOnWhenIdle=true address is the ZigBee
    # compliant destination for multicasts. 0=false, 1=true
    vol.Optional(
        EzspConfigId.CONFIG_SEND_MULTICASTS_TO_SLEEPY_ADDRESS.name
    ): cv_optional_int(min=0, max=1),
    #
    # ZLL group address initial configuration
    vol.Optional(EzspConfigId.CONFIG_ZLL_GROUP_ADDRESSES.name): cv_optional_int(
        min=0, max=255
    ),
    #
    # ZLL RSSI threshold initial configuration
    vol.Optional(EzspConfigId.CONFIG_ZLL_RSSI_THRESHOLD.name): cv_optional_int(
        min=-128, max=127
    ),
    #
    # The maximum number of pairings supported by the stack. Controllers must
    # support at least one pairing table entry while targets must support at
    # least five
    vol.Optional(EzspConfigId.CONFIG_RF4CE_PAIRING_TABLE_SIZE.name): cv_optional_int(
        min=0, max=126
    ),
    #
    # The maximum number of outgoing RF4CE packets supported by the stack
    vol.Optional(
        EzspConfigId.CONFIG_RF4CE_PENDING_OUTGOING_PACKET_TABLE_SIZE.name
    ): cv_optional_int(min=0, max=16),
    #
    # Toggles the MTORR flow control in the stack. 0=false, 1=true
    vol.Optional(EzspConfigId.CONFIG_MTORR_FLOW_CONTROL.name): cv_optional_int(
        min=0, max=1
    ),
    #
    # Deprecated
    # The amount of time a trust center will store a transient key
    # with which a device can use to join the network
    vol.Optional(EzspConfigId.CONFIG_TRANSIENT_KEY_TIMEOUT_S.name): cv_optional_int(
        min=0, max=65535
    ),
}

EZSP_POLICIES_SHARED = {
    vol.Optional(
        EzspPolicyId.TC_KEY_REQUEST_POLICY.name,
        default=EzspDecisionId.ALLOW_TC_KEY_REQUESTS,
    ): cv_uint16,
    vol.Optional(
        EzspPolicyId.APP_KEY_REQUEST_POLICY.name,
        default=EzspDecisionId.DENY_APP_KEY_REQUESTS,
    ): cv_uint16,
    vol.Optional(
        EzspPolicyId.TRUST_CENTER_POLICY.name,
        default=EzspDecisionId.ALLOW_PRECONFIGURED_KEY_JOINS,
    ): cv_uint16,
}

EZSP_POLICIES_SCH = {
    **EZSP_POLICIES_SHARED,
    **{vol.Optional(policy.name): cv_uint16 for policy in EzspPolicyId},
}



"""Protocol version 10 specific structs."""

from __future__ import annotations

from bellows.ezsp.v9.types.struct import (  # noqa: F401
    EmberBeaconClassificationParams,
    EmberBeaconData,
    EmberBeaconIterator,
    EmberDutyCycleLimits,
    EmberGpProxyTableEntry,
    EmberGpSinkListEntry,
    EmberGpSinkTableEntry,
    EmberKeyStruct,
    EmberPerDeviceDutyCycle,
    EmberTransientKeyData,
)
import bellows.types.basic as basic
from bellows.types.struct import (  # noqa: F401
    EmberAesMmoHashContext,
    EmberApsFrame,
    EmberBindingTableEntry,
    EmberCurrentSecurityState,
    EmberGpAddress,
    EmberInitialSecurityState,
    EmberMulticastTableEntry,
    EmberNeighborTableEntry,
    EmberNetworkParameters,
    EmberRouteTableEntry,
    EmberTokenData,
    EmberTokenInfo,
    EmberTokTypeStackZllData,
    EmberTokTypeStackZllSecurity,
    EmberZigbeeNetwork,
    EmberZllAddressAssignment,
    EmberZllDeviceInfoRecord,
    EmberZllInitialSecurityState,
    EmberZllNetwork,
    EmberZllSecurityAlgorithmData,
    EzspStruct,
)

from . import named


class EmberChildData(EzspStruct):
    """A structure containing a child node's data."""

    # The EUI64 of the child
    eui64: named.EUI64
    # The node type of the child
    type: named.EmberNodeType
    # The short address of the child
    id: named.EmberNodeId
    # The phy of the child
    phy: basic.uint8_t
    # The power of the child
    power: basic.uint8_t
    # The timeout of the child
    timeout: basic.uint8_t
    timeout_remaining: basic.uint32_t

    # The GPD's EUI64.
    # gpdIeeeAddress: named.EUI64
    # The GPD's source ID.
    # sourceId: basic.uint32_t



from . import types as t
from ..v10.commands import COMMANDS as COMMANDS_v10

COMMANDS = {
    **COMMANDS_v10,
    "pollHandler": (
        0x0044,
        (),
        tuple({"childId": t.EmberNodeId, "transmitExpected": t.Bool}.values()),
    ),
}



"""Protocol version 13 specific structs."""

from __future__ import annotations

from bellows.ezsp.v12.types.struct import *  # noqa: F401, F403
from bellows.ezsp.v13.types import named
import bellows.types as t
from bellows.types.struct import EzspStruct


class sl_zb_sec_man_context_t(EzspStruct):
    """Context for Zigbee Security Manager operations."""

    # The type of key being referenced.
    core_key_type: named.sl_zb_sec_man_key_type_t
    # The index of the referenced key.
    key_index: t.uint8_t
    # The type of key derivation operation to perform on a key.
    derived_type: named.sl_zb_sec_man_derived_key_type_t
    # The EUI64 associated with this key.
    eui64: t.EUI64
    # Multi-network index.
    multi_network_index: t.uint8_t
    # Flag bitmask.
    flags: named.sl_zb_sec_man_flags_t
    # Algorithm to use with this key (for PSA APIs)
    psa_key_alg_permission: t.uint32_t


class sl_zb_sec_man_network_key_info_t(EzspStruct):
    """The metadata pertaining to an network key."""

    network_key_set: t.Bool
    alternate_network_key_set: t.Bool
    network_key_sequence_number: t.uint8_t
    alt_network_key_sequence_number: t.uint8_t
    network_key_frame_counter: t.uint32_t



from . import types as t
from ..v12.commands import COMMANDS as COMMANDS_v12

COMMANDS = {
    **COMMANDS_v12,
    "getNetworkKeyInfo": (
        0x0116,
        (),
        tuple(
            {
                "status": t.sl_Status,
                "network_key_info": t.sl_zb_sec_man_network_key_info_t,
            }.values()
        ),
    ),
    "gpSecurityTestVectors": (
        0x0117,
        (),
        tuple(
            {
                "status": t.EmberStatus,
            }.values()
        ),
    ),
    "tokenFactoryReset": (
        0x0077,
        tuple({"excludeOutgoingFC": t.Bool, "excludeBootCounter": t.Bool}.values()),
        (),
    ),
    "gpSinkTableGetNumberOfActiveEntries": (
        0x0118,
        (),
        tuple(
            {
                "number_of_entries": t.uint8_t,
            }.values()
        ),
    ),
    # The following commands are redefined because `sl_zb_sec_man_context_t` changed
    "exportKey": (
        0x0114,
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
            }.values()
        ),
        tuple(
            {
                "key": t.KeyData,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportTransientKeyByEui": (
        0x0113,
        tuple(
            {
                "eui64": t.EUI64,
            }.values()
        ),
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "plaintext_key": t.KeyData,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "exportTransientKeyByIndex": (
        0x0112,
        tuple(
            {
                "index": t.uint8_t,
            }.values()
        ),
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "plaintext_key": t.KeyData,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "getApsKeyInfo": (
        0x010C,
        tuple(
            {
                "context_in": t.sl_zb_sec_man_context_t,
            }.values()
        ),
        tuple(
            {
                "eui": t.EUI64,
                "key_data": t.sl_zb_sec_man_aps_key_metadata_t,
                "status": t.sl_Status,
            }.values()
        ),
    ),
    "importKey": (
        0x0115,
        tuple(
            {
                "context": t.sl_zb_sec_man_context_t,
                "key": t.KeyData,
            }.values()
        ),
        tuple(
            {
                "status": t.sl_Status,
            }.values()
        ),
    ),
}

del COMMANDS["becomeTrustCenter"]  # this one was likely removed earlier
del COMMANDS["getKey"]
del COMMANDS["getKeyTableEntry"]
del COMMANDS["setKeyTableEntry"]
del COMMANDS["addOrUpdateKeyTableEntry"]
del COMMANDS["addTransientLinkKey"]
del COMMANDS["getTransientLinkKey"]
del COMMANDS["getTransientKeyTableEntry"]
del COMMANDS["setSecurityKey"]
del COMMANDS["setSecurityParameters"]
del COMMANDS["resetToFactoryDefaults"]
del COMMANDS["getSecurityKeyStatus"]
